from scoop._control import _stat
import unittest

class TestStat(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestStat, self).__init__(*args, **kwargs)

    def test_appendleft(self):
        stats = _stat()
        data = list(range(5))
        for i in data[::-1]:
            stats.appendleft(i)
        for i, j in enumerate(stats):
            self.assertEqual(data[i], j)
     
    def test_maxlen(self):
        stats = _stat()
        data = list(range(15))
        for i in data[::-1]:
            stats.appendleft(i)
        for i, j in enumerate(stats):
            self.assertEqual(data[i], j)
        self.assertEqual(len(stats), 10)

    def test_mean(self):
        stats = _stat()
        data = list(range(15))
        for i in data:
            stats.appendleft(float(i))
        self.assertEqual(stats.mean(), 9.5)
        stats.appendleft(1000)
        self.assertEqual(stats.mean(), 109.0)

    def test_std(self):
        stats = _stat()
        data = list(range(10))
        for i in data:
            stats.appendleft(float(i))
        self.assertAlmostEqual(stats.std(), 2.87228132327)
        stats.appendleft(1000)
        self.assertAlmostEqual(stats.std(), 298.510050082)

if __name__ == "__main__":
    t = unittest.TestLoader().loadTestsFromTestCase(TestStat)
    unittest.TextTestRunner(verbosity=2).run(t)



#!/usr/bin/env python
#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#

"""This examples imports a list of 100 web sites and returns
the sizes in bytes of every web site. It compares the speed
to do this task of the regular python map, futures.map and
the as_completed function."""

import urllib.request
import urllib.error
from scoop import futures
import socket
import time

def getSize(string):
    """ This functions opens a web sites and then calculate the total
    size of the page in bytes. This is for the sake of the example. Do
    not use this technique in real code as it is not a very bright way
    to do this."""
    try:
        # We open the web page
        with urllib.request.urlopen(string, None, 1) as f:
            return sum(len(line) for line in f)
    except (urllib.error.URLError, socket.timeout) as e:
        return 0

if __name__ == "__main__":
    # The pageurl variable contains a link to a list of web sites. It is
    # commented for security's sake.
    pageurl = "http://httparchive.org/lists/Fortune%20500.txt"
    #pageurl  = "http://www.example.com"
    with urllib.request.urlopen(pageurl) as pagelist:
        pages = [page.decode() for page in pagelist][:30]

    # This will apply the getSize function on every item of the pages list
    # in parallel. The results will be treated as they are produced.
    fut = [futures.submit(getSize, page) for page in pages]
    for f in futures.as_completed(fut):
        time.sleep(0.1) # Work on the data
        print(f.result())



#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
"""Module containing builins fallbacks or exceptions when SCOOP is not
started properly."""
import sys
import warnings
from functools import wraps


class NotStartedProperly(Exception):
    """SCOOP was not started properly"""
    pass


def ensureScoopStartedProperlyMapFallback(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        futures_not_loaded = 'scoop.futures' not in sys.modules
        controller_not_started = not (
            sys.modules['scoop.futures'].__dict__.get("_controller", None)
        )
        if futures_not_loaded or controller_not_started:
            if not hasattr(ensureScoopStartedProperlyMapFallback, "already"):
                warnings.warn(
                    "SCOOP was not started properly.\n"
                    "Be sure to start your program with the "
                    "'-m scoop' parameter. You can find "
                    "further information in the "
                    "documentation.\n"
                    "Your map call has been replaced by the builtin "
                    "serial Python map().",
                    RuntimeWarning
                )
                ensureScoopStartedProperlyMapFallback.already = True
            return map(*args, **kwargs)
        return func(*args, **kwargs)
    return wrapper


def ensureScoopStartedProperly(func):
    def wrapper(*args, **kwargs):
        futures_not_loaded = 'scoop.futures' not in sys.modules
        controller_not_started = not (
            sys.modules['scoop.futures'].__dict__.get("_controller", None)
        )
        if futures_not_loaded or controller_not_started:
            raise NotStartedProperly("SCOOP was not started properly.\n"
                                     "Be sure to start your program with the "
                                     "'-m scoop' parameter. You can find "
                                     "further information in the "
                                     "documentation.")
        return func(*args, **kwargs)
    return wrapper


#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
"""
Calculation of Pi using a Monte Carlo method. This version is used in the 
documentation and should be kept as clean as possible for study purposes.
"""

from math import hypot
from random import random
from scoop import futures

def test(tries):
    return sum(hypot(random(), random()) < 1 for _ in range(tries))

def calcPi(nbFutures, tries):
    expr = futures.map(test, [tries] * nbFutures)
    return 4. * sum(expr) / float(nbFutures * tries)

if __name__ == "__main__":
    print("pi = {}".format(calcPi(3000, 5000)))



#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
"""
Computing a residual sum of squares using SCOOP.
"""

from __future__ import print_function
import random
import operator
import time
from scoop import futures

# The data is generated on every workers for the sake of this example.
# This data could come from a file located on a shared hard drive or else.
random.seed(31415926)
leftSignal = [random.randint(-100, 100) for _ in range(200000)]
rightSignal = [random.randint(-100, 100) for _ in range(200000)]

# Set the size of each worker batch
PARALLEL_SIZE = 25000

# Set the parallel function that will compute the Residual Sum of Squares
# The index represent the element 
def RSS(index):
    # Get the data interval to compute on a given Future
    data = zip(leftSignal[index:index+PARALLEL_SIZE],
               rightSignal[index:index+PARALLEL_SIZE])
    return sum(abs(y - x)**2 for y, x in data)


if __name__ == "__main__":
    # Parallel with reduction call
    # Take a beginning timestamp
    ts = time.time()
    # Generate indexes to pass to futures
    indexes = range(0,
                    len(leftSignal),
                    PARALLEL_SIZE,
                    )
    # Execute the RSS computation parallely
    presult = futures.mapReduce(RSS,
                                operator.add,
                                indexes,
                                )
    ptime = time.time() - ts
    print("mapReduce result obtained in {0:03f}s".format(ptime))

    # Serial
    ts = time.time()
    sresult = sum(abs(a - b)**2 for a, b in zip(leftSignal, rightSignal))
    stime = time.time() - ts
    print("Serial result obtained in {0:03f}s".format(stime))
    assert presult == sresult



#!/usr/bin/env python
#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
from threading import Thread
import subprocess
import shlex
import sys
import os
try:
    import cPickle as pickle
except ImportError:
    import pickle

import scoop
try:
    import psutil
except ImportError:
    psutil = None


class localBroker(object):
    def __init__(self, debug, nice=0, backend='ZMQ'):
        """Starts a broker on random unoccupied ports"""
        self.backend = backend
        if backend == 'ZMQ':
            from ..broker.brokerzmq import Broker
        else:
            from ..broker.brokertcp import Broker
        if nice:
            if not psutil:
                scoop.logger.error("'nice' used while psutil not installed.")
                raise ImportError("psutil is needed for nice functionnality.")
            p = psutil.Process(os.getpid())
            p.set_nice(nice)
        self.localBroker = Broker(debug=debug)
        self.brokerPort, self.infoPort = self.localBroker.getPorts()
        self.broker = Thread(target=self.localBroker.run)
        self.broker.daemon = True
        self.broker.start()
        scoop.logger.debug("Local broker launched on ports {0}, {1}"
                      ".".format(self.brokerPort, self.infoPort))

    def sendConnect(self, data):
        """Send a CONNECT command to the broker
            :param data: List of other broker main socket URL"""
        # Imported dynamically - Not used if only one broker
        if self.backend == 'ZMQ':
            import zmq
            self.context = zmq.Context()
            self.socket = self.context.socket(zmq.DEALER)
            self.socket.setsockopt(zmq.IDENTITY, b'launcher')
            self.socket.connect(
                "tcp://127.0.0.1:{port}".format(
                    port=self.brokerPort,
                )
            )
            self.socket.send_multipart([b"CONNECT",
                                        pickle.dumps(data,
                                                     pickle.HIGHEST_PROTOCOL)])
        else:
            # TODO
            pass

    def getHost(self):
        return "127.0.0.1"

    def getPorts(self):
        return (self.brokerPort, self.infoPort)

    def close(self):
        scoop.logger.debug('Closing local broker.')


class remoteBroker(object):
    BASE_SSH = ['ssh', '-x', '-n', '-oStrictHostKeyChecking=no']

    def __init__(self, hostname, pythonExecutable, debug=False, nice=0,
                 backend='ZMQ'):
        """Starts a broker on the specified hostname on unoccupied ports"""
        self.backend = backend
        brokerString = ("{pythonExec} -m scoop.broker.__main__ "
                        "--echoGroup "
                        "--echoPorts "
                        "--backend {backend} ".format(
                            pythonExec=pythonExecutable,
                            backend=backend,
                            )
                        )
        if nice:
            brokerString += "--nice {nice} ".format(nice=nice)
        if debug:
            brokerString += "--debug --path {path} ".format(
                path=os.getcwd()
            )
        self.hostname = hostname
        for i in range(5000, 10000, 2):
            cmd = self.BASE_SSH + [
                hostname,
                brokerString.format(brokerPort=i,
                                    infoPort=i + 1,
                                    pythonExec=pythonExecutable,
                                    )
            ]
            scoop.logger.debug("Launching remote broker: {cmd}"
                               "".format(cmd=" ".join(cmd)))
            self.shell = subprocess.Popen(cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            # TODO: This condition is not doing what it's supposed
            if self.shell.poll() is not None:
                continue
            else:
                self.brokerPort, self.infoPort = i, i + 1
                break
        else:
            raise Exception("Could not successfully launch the remote broker.")

        # Get remote process group ID
        try:
            self.remoteProcessGID = int(self.shell.stdout.readline().strip())
        except ValueError:
            self.remoteProcessGID = None

        # Get remote ports
        receivedLine = self.shell.stdout.readline()
        try:
            ports = receivedLine.decode().strip().split(",")
            self.brokerPort, self.infoPort = ports
        except ValueError:
            # Following line for Python 2.6 compatibility (instead of [as e])
            e = sys.exc_info()[1]

            # Terminate the process, otherwide reading from stderr may wait
            # undefinitely
            self.shell.terminate()

            stderr = self.shell.stderr.read()
            raise Exception("Could not successfully launch the remote broker.\n"
                            "Requested remote broker ports, received:\n"
                            "{receivedLine}\n"
                            "Port number decoding error:\n{e}\n"
                            "SSH process stderr:\n{stderr}".format(**locals()))

        scoop.logger.debug("Foreign broker launched on ports {0}, {1} of host {2}"
                      ".".format(self.brokerPort,
                                 self.infoPort,
                                 hostname,
                                 )
                      )

    def sendConnect(self, data):
        """Send a CONNECT command to the broker
            :param data: List of other broker main socket URL"""
        # Imported dynamically - Not used if only one broker
        if self.backend == 'ZMQ':
            import zmq
            self.context = zmq.Context()
            self.socket = self.context.socket(zmq.DEALER)
            if sys.version_info < (3,):
                self.socket.setsockopt_string(zmq.IDENTITY, unicode('launcher'))
            else:
                self.socket.setsockopt_string(zmq.IDENTITY, 'launcher')
            self.socket.connect(
                "tcp://{hostname}:{port}".format(
                    port=self.brokerPort,
                    hostname = self.hostname
                )
            )
            self.socket.send_multipart([b"CONNECT",
                                        pickle.dumps(data,
                                                     pickle.HIGHEST_PROTOCOL)])
        else:
            # TODO
            pass


    def getHost(self):
        return self.hostname

    def getPorts(self):
        return (self.brokerPort, self.infoPort)

    def isLocal(self):
        """Is the current broker on the localhost?"""
        # This exists for further fusion with localBroker
        return False
        # return self.hostname in utils.localHostnames

    def close(self):
        """Connection(s) cleanup."""
        # TODO: DRY with workerLaunch.py
        # Ensure everything is cleaned up on exit
        scoop.logger.debug('Closing broker on host {0}.'.format(self.hostname))

        # Terminate subprocesses
        try:
            self.shell.terminate()
        except OSError:
            pass

        # Send termination signal to remaining workers
        if not self.isLocal() and self.remoteProcessGID is None:
            scoop.logger.warn(
                "Zombie process(es) possibly left on host {0}!"
                "".format(self.hostname)
            )
        elif not self.isLocal():
            command = ("python -c "
                       "'import os, signal; os.killpg({0}, signal.SIGKILL)' "
                       ">&/dev/null").format(self.remoteProcessGID)
            subprocess.Popen(self.BASE_SSH
                             + [self.hostname]
                             + [command],
            ).wait()

        # Output child processes stdout and stderr to console
        sys.stdout.write(self.shell.stdout.read().decode("utf-8"))
        sys.stdout.flush()

        sys.stderr.write(self.shell.stderr.read().decode("utf-8"))
        sys.stderr.flush()



#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
# Global imports
from collections import namedtuple
import logging
import sys
import subprocess

# Local
import scoop
from scoop import utils


class Host(object):
    """Represents an accessible computing resource.
       Can be remote (ssh via netowrk) or represent localhost."""
    BOOTSTRAP_MODULE = 'scoop.bootstrap.__main__'
    BASE_SSH = ['ssh', '-x', '-n', '-oStrictHostKeyChecking=no']
    LAUNCHING_ARGUMENTS = namedtuple(
        'launchingArguments',
        [
            'pythonPath', 'path', 'nice', 'pythonExecutable', 'size', 'origin',
            'brokerHostname', 'brokerPorts', 'debug', 'profiling', 'executable',
            'verbose', 'args', 'prolog', 'backend'
        ]
    )

    def __init__(self, hostname="localhost"):
        self.workersArguments = []
        self.hostname = hostname
        self.subprocesses = []
        self.remoteProcessGID = None

    def __repr__(self):
        return "{0} ({1} workers)".format(
            self.hostname,
            len(self.workersArguments)
        )

    def isLocal(self):
        """Is the current host the localhost?"""
        return self.hostname in utils.localHostnames

    def addWorker(self, *args, **kwargs):
        """Add a worker assignation
            arguments and order to pass are defined in LAUNCHING_ARGUMENTS
            Using named args is advised.
        """
        try:
            la = self.LAUNCHING_ARGUMENTS(*args, **kwargs)
        except TypeError as e:
            scoop.logger.error(("addWorker failed to convert args %s and kwargs %s "
                            "to namedtuple (requires %s arguments (names %s)") %
                            (args, kwargs, len(self.LAUNCHING_ARGUMENTS._fields),
                             self.LAUNCHING_ARGUMENTS._fields))

        self.workersArguments.append(la)

    def _WorkerCommand_environment(self, worker):
        """Return list of shell commands to prepare the environment for bootstrap"""
        c = []
        if worker.prolog:
            c.extend([
                "source",
                worker.prolog,
                "&&",
            ])
        if worker.pythonPath:
            # Tried to make it compliant to all shell variants.
            c.extend([
                "export",
                "PYTHONPATH={0}:$PYTHONPATH".format(worker.pythonPath),
                ">&/dev/null",
                "||",
                "setenv",
                "PYTHONPATH",
                "{0}:$PYTHONPATH".format(worker.pythonPath),
                "&&",
            ])
        return c

    def _WorkerCommand_bootstrap(self, worker):
        """Return list commands to start the bootstrap process"""
        c = []
        c.extend([worker.pythonExecutable, '-m', self.BOOTSTRAP_MODULE])
        return c

    def _WorkerCommand_options(self, worker, workerID):
        """Return list of options for bootstrap"""
        c = []
        # If broker is on localhost
        if self.hostname == worker.brokerHostname:
            broker = "127.0.0.1"
        else:
            broker = worker.brokerHostname

        # If host is not localhost, echo group process
        if not self.isLocal() and workerID == 0:
            c.append("--echoGroup ")

        if worker.nice is not None:
            c.extend(['--nice', str(worker.nice)])
        c.extend(['--size', str(worker.size)])
        c.extend(['--workingDirectory', str(worker.path)])
        c.extend(['--brokerHostname', broker])
        c.extend(['--externalBrokerHostname', worker.brokerHostname])
        c.extend(['--taskPort', str(worker.brokerPorts[0])])
        c.extend(['--metaPort', str(worker.brokerPorts[1])])
        if worker.origin and worker.executable:
            c.append('--origin')
        if worker.debug:
            c.append('--debug')
        if worker.profiling:
            c.append('--profile')
        if worker.backend:
            c.append('--backend={0}'.format(worker.backend))
        if worker.verbose == 0:
            c.append('-q')
        elif worker.verbose >= 2:
            c.append('-v')
        return c

    def _WorkerCommand_executable(self, worker):
        """Return executable and any options to be executed by bootstrap"""
        c = []
        if worker.executable:
            c.append(worker.executable)
        # This trick is used to parse correctly quotes
        # (ie. myScript.py 'arg1 "arg2" arg3')
        # Because shell=True is set with Popen, every quote gets re-interpreted
        # It replaces simple quotation marks with \\\" which gets evaluated to
        # \" by the second shell which prints it out as a double quote.
        if worker.args:
            if self.isLocal():
                # If local, no shell is used so no escaping needed
                c.extend([
                    '{0}'.format(a) for a in worker.args
                ])
            else:
                c.extend([
                    '"{0}"'.format(a.replace('"', '\\\"')) for a in worker.args
                ])
        return c

    def _getWorkerCommandList(self, workerID):
        """Generate the workerCommand as list"""
        worker = self.workersArguments[workerID]

        c = []
        if not self.isLocal():
            c.extend(self._WorkerCommand_environment(worker))

        c.extend(self._WorkerCommand_bootstrap(worker))
        c.extend(self._WorkerCommand_options(worker, workerID))
        c.extend(self._WorkerCommand_executable(worker))

        return c

    def getWorkerCommand(self, workerID=None):
        """Retrieves the working launching shell command."""
        c = " ".join(self._getWorkerCommandList(workerID))
        return c

    def getCommand(self):
        """Retrieves the shell command to launch every worker on this host."""
        # All this parenthesis insanity is to start subshells (workers) in the
        # correct monitoring mode (no background job echo).
        # Output: ( [launch command 1] & ) && ( [launch command 2] & ) [...]
        command = []
        for workerID, worker in enumerate(self.workersArguments):
            command.append("(" + self.getWorkerCommand(workerID))
        command[-1] += ")"
        return " & ) && ".join(command)

    def launch(self, tunnelPorts=None, stdPipe=False):
        """Launch every worker assigned on this host."""
        if self.isLocal():
            # Launching local workers
            for workerID, workerToLaunch in enumerate(self.workersArguments):
                # Launch one per subprocess
                c = self._getWorkerCommandList(workerID)
                self.subprocesses.append(
                    subprocess.Popen(c)
                )
        else:
            # Launching remotely
            sshCmd = self.BASE_SSH
            if tunnelPorts is not None:
                sshCmd += [
                    '-R {0}:127.0.0.1:{0}'.format(tunnelPorts[0]),
                    '-R {0}:127.0.0.1:{0}'.format(tunnelPorts[1]),
                ]
            self.subprocesses.append(
                subprocess.Popen(sshCmd + [self.hostname, self.getCommand()],
                                 stdout=subprocess.PIPE if stdPipe else None,
                                 stderr=subprocess.PIPE if stdPipe else None,
                )
            )
            # Get group id from remote connections
            receivedLine = self.subprocesses[-1].stdout.readline()
            try:
                textGID = receivedLine.decode().strip()
                self.remoteProcessGID = int(textGID)
            except ValueError:
                # Following line for Python 2.6 compatibility (instead of [as e])
                e = sys.exc_info()[1]

                # Terminate the process, otherwide reading from stderr may wait
                # undefinitely
                self.subprocesses[-1].terminate()

                stderr = self.subprocesses[-1].stderr.read()
                hostname = self.hostname
                scoop.logger.warning("Could not successfully launch the remote "
                                 "worker on {hostname}.\n"
                                 "Requested remote group process id, "
                                 "received:\n{receivedLine}\n"
                                 "Group id decoding error:\n{e}\n"
                                 "SSH process stderr:\n{stderr}"
                                 "".format(**locals()))

        return self.subprocesses

    def close(self):
        """Connection(s) cleanup."""
        # Ensure everything is cleaned up on exit
        scoop.logger.debug('Closing workers on {0}.'.format(self))

        # Output child processes stdout and stderr to console
        for process in self.subprocesses:
            if process.stdout is not None:
                sys.stdout.write(process.stdout.read().decode("utf-8"))
                sys.stdout.flush()

            if process.stderr is not None:
                sys.stderr.write(process.stderr.read().decode("utf-8"))
                sys.stderr.flush()

        # Terminate subprocesses
        for process in self.subprocesses:
            try:
                process.terminate()
            except OSError:
                pass

        # Send termination signal to remaining workers
        if not self.isLocal() and self.remoteProcessGID is None:
                scoop.logger.warn("Zombie process(es) possibly left on "
                             "host {0}!".format(self.hostname))
        elif not self.isLocal():
            command = ("python -c "
                       "'import os, signal; os.killpg({0}, signal.SIGKILL)' "
                       ">&/dev/null").format(self.remoteProcessGID)
            subprocess.Popen(self.BASE_SSH
                             + [self.hostname]
                             + [command],
            ).wait()



from scoop import futures, shared
import scoop
import code

# Usage example:
# 
# >>> def myExample(inVal):
# ...     return inVal + 1
# ...
# >>> shared.setConst(myExample=myExample)
# >>> print(list(futures.map(myExample, range(64))))
# >>> exit()

if __name__ == '__main__':
    code.interact(local=locals())


#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
"""
Example of dynamic parallel object manipulation.
"""
from scoop import futures


class myClass(object):
    """An object with an instance variable."""
    def __init__(self):
        self.myVar = 5


def modifyClass(myInstance):
    """Function modifying an instance variable."""
    myInstance.myVar += 1
    return myInstance


def main():
    # Create object instances
    myInstances = [myClass() for _ in range(20)]
    # Modify them parallely
    myAnswers = list(futures.map(modifyClass, myInstances))

    # Each result is a new object with the modifications applied
    print(myAnswers)
    print([a.myVar for a in myAnswers])


if __name__ == "__main__":
    main()



from scoop._types import StopWatch

import unittest
import time

class TestStopWatch(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestStopWatch, self).__init__(*args, **kwargs)

    def test_get(self):
        watch = StopWatch()
        first = watch.get()
        time.sleep(0.1)
        second = watch.get()
        # *nix tend to overshoot a tiny bit, Windows tend to always be under
        # by max. 1ms
        self.assertAlmostEqual(second - first, 0.1, places=2)

    def test_halt(self):
        watch = StopWatch()
        watch.halt()
        first = watch.get()
        time.sleep(0.1)
        second = watch.get()
        self.assertEqual(first, second)

    def test_resume(self):
        watch = StopWatch()
        watch.halt()
        first = watch.get()
        watch.resume()
        time.sleep(0.1)
        second = watch.get()
        # See test_get
        self.assertAlmostEqual(second - first, 0.1, places=2)

    def test_reset(self):
        watch = StopWatch()
        time.sleep(0.1)
        watch.reset()
        self.assertLess(watch.get(), 0.001)


if __name__ == "__main__":
    t = unittest.TestLoader().loadTestsFromTestCase(TestStopWatch)
    unittest.TextTestRunner(verbosity=2).run(t)



#    This file is part of DEAP.
#
#    DEAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    DEAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with DEAP. If not, see <http://www.gnu.org/licenses/>.

try:
    from itertools import product
except ImportError:
    def product(*args, **kwds):
        # product('ABCD', 'xy') --> Ax Ay Bx By Cx Cy Dx Dy
        # product(range(2), repeat=3) --> 000 001 010 011 100 101 110 111
        pools = map(tuple, args) * kwds.get('repeat', 1)
        result = [[]]
        for pool in pools:
            result = [x+[y] for x in result for y in pool]
        for prod in result:
            yield tuple(prod)
            
class SortingNetwork(list):
    """Sorting network class.
    
    From Wikipedia : A sorting network is an abstract mathematical model
    of a network of wires and comparator modules that is used to sort a
    sequence of numbers. Each comparator connects two wires and sort the
    values by outputting the smaller value to one wire, and a larger
    value to the other.
    """
    def __init__(self, dimension, connectors = []):
        self.dimension = dimension
        for wire1, wire2 in connectors:
            self.addConnector(wire1, wire2)
    
    def addConnector(self, wire1, wire2):
        """Add a connector between wire1 and wire2 in the network."""
        if wire1 == wire2:
            return
        
        if wire1 > wire2:
            wire1, wire2 = wire2, wire1
        
        try:
            last_level = self[-1]
        except IndexError:
            # Empty network, create new level and connector
            self.append([(wire1, wire2)])
            return
        
        for wires in last_level:
            if wires[1] >= wire1 and wires[0] <= wire2:
                self.append([(wire1, wire2)])
                return
        
        last_level.append((wire1, wire2))
    
    def sort(self, values):
        """Sort the values in-place based on the connectors in the network."""
        for level in self:
            for wire1, wire2 in level:
                if values[wire1] > values[wire2]:
                    values[wire1], values[wire2] = values[wire2], values[wire1]
    
    def assess(self, cases=None):
        """Try to sort the **cases** using the network, return the number of
        misses. If **cases** is None, test all possible cases according to
        the network dimensionality.
        """
        if cases is None:
            cases = product(range(2), repeat=self.dimension)
        
        misses = 0
        ordered = [[0]*(self.dimension-i) + [1]*i for i in range(self.dimension+1)]
        for sequence in cases:
            sequence = list(sequence)
            self.sort(sequence)
            misses += (sequence != ordered[sum(sequence)])
        return misses
    
    def draw(self):
        """Return an ASCII representation of the network."""
        str_wires = [["-"]*7 * self.depth]
        str_wires[0][0] = "0"
        str_wires[0][1] = " o"
        str_spaces = []

        for i in range(1, self.dimension):
            str_wires.append(["-"]*7 * self.depth)
            str_spaces.append([" "]*7 * self.depth)
            str_wires[i][0] = str(i)
            str_wires[i][1] = " o"
        
        for index, level in enumerate(self):
            for wire1, wire2 in level:
                str_wires[wire1][(index+1)*6] = "x"
                str_wires[wire2][(index+1)*6] = "x"
                for i in range(wire1, wire2):
                    str_spaces[i][(index+1)*6+1] = "|"
                for i in range(wire1+1, wire2):
                    str_wires[i][(index+1)*6] = "|"
        
        network_draw = "".join(str_wires[0])
        for line, space in zip(str_wires[1:], str_spaces):
            network_draw += "\n"
            network_draw += "".join(space)
            network_draw += "\n"
            network_draw += "".join(line)
        return network_draw
    
    @property
    def depth(self):
        """Return the number of parallel steps that it takes to sort any input.
        """
        return len(self)
    
    @property
    def length(self):
        """Return the number of comparison-swap used."""
        return sum(len(level) for level in self)




#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
"""
A simple example showing how to resolve a full balanced tree with multiples
techniques using SCOOP.
"""

from scoop import futures

def func4(n):
    # Example of calculus you may want to perform
    result = n*n
    return result

def func3(n):
    # This call results in a generator function
    result = futures.map(func4, [i+1 for i in range(n)])
    # The results are evaluated here when they are accessed.
    return sum(result)

def func2(n):
    launches = [futures.submit(func3, i + 1) for i in range(n)]
    # Spawn a generator for each completion, unordered
    result = (a.result() for a in futures.as_completed(launches))
    return sum(result)

def func1(n):
    # To force an immediate evaluation, you can wrap your map in a list such as:
    result = list(futures.map(func2, [i+1 for i in range(n)]))
    return sum(result)

def func0(n):
    # Task submission is asynchronous; It will return immediately.
    task = futures.submit(func1, n)
    # The call blocks here until it gets the result
    result = task.result()
    return result

def main():
    task = futures.submit(func0, 20)
    # You can wait for a result before continuing computing
    futures.wait([task], return_when=futures.ALL_COMPLETED)
    result = task.result()
    print(result)
    return result

if __name__ == "__main__":
    main()



#!/usr/bin/env python
#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP). 
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#

import itertools
from inspect import ismethod
from functools import reduce
import time

from . import encapsulation, utils
import scoop
from .fallbacks import ensureScoopStartedProperly, NotStartedProperly


elements = None


def _ensureAtomicity(fn):
    """Ensure atomicity of passed elements on the whole worker pool"""
    @ensureScoopStartedProperly
    def wrapper(*args, **kwargs):
        """setConst(**kwargs)
        Set a constant that will be shared to every workers.
        This call blocks until the constant has propagated to at least one
        worker.

        :param \*\*kwargs: One or more combination(s) key=value. Key being the
            variable name and value the object to share.

        :returns: None.

        Usage: setConst(name=value)
        """
        # Note that the docstring is the one of setConst.
        # This is because of the documentation framework (sphinx) limitations.

        from . import _control

        # Enforce retrieval of currently awaiting constants
        _control.execQueue.socket.pumpInfoSocket()

        for key, value in kwargs.items():
            # Object name existence check
            if key in itertools.chain(*(elem.keys() for elem in elements.values())):
                raise TypeError("This constant already exists: {0}.".format(key))

        # Retry element propagation until it is returned
        while all(key in elements.get(scoop.worker, []) for key in kwargs.keys()) is not True:
            scoop.logger.debug("Sending global variables {0}...".format(
                list(kwargs.keys())
            ))
            # Call the function
            fn(*args, **kwargs)

            # Enforce retrieval of currently awaiting constants
            _control.execQueue.socket.pumpInfoSocket()

            # TODO: Make previous blocking instead of sleep
            time.sleep(0.1)

        # Atomicity check
        elementNames = list(itertools.chain(*(elem.keys() for elem in elements.values())))
        if len(elementNames) != len(set(elementNames)):
            raise TypeError("This constant already exists: {0}.".format(key))

    return wrapper


@_ensureAtomicity
def setConst(**kwargs):
    """setConst(**kwargs)
    Set a constant that will be shared to every workers.

    :param **kwargs: One or more combination(s) key=value. Key being the
        variable name and value the object to share.

    :returns: None.

    Usage: setConst(name=value)
    """
    from . import _control
    
    sendVariable = _control.execQueue.socket.sendVariable

    for key, value in kwargs.items():
        # Propagate the constant
        # for file-like objects, see encapsulation.py where copyreg was
        # used to overload standard pickling.
        if callable(value):
            sendVariable(key, encapsulation.FunctionEncapsulation(value, key))
        else:
            sendVariable(key, value)


def getConst(name, timeout=0.1):
    """Get a shared constant.

    :param name: The name of the shared variable to retrieve.
    :param timeout: The maximum time to wait in seconds for the propagation of
        the constant.

    :returns: The shared object.

    Usage: value = getConst('name')
    """
    from . import _control
    import time

    timeStamp = time.time()
    while True:
        # Enforce retrieval of currently awaiting constants
        _control.execQueue.socket.pumpInfoSocket()

        # Constants concatenation
        constants = dict(reduce(
            lambda x, y: x + list(y.items()),
            elements.values(),
            []
        ))
        timeoutHappened = time.time() - timeStamp > timeout
        if constants.get(name) is not None or timeoutHappened:
            return constants.get(name)
        time.sleep(0.01)


class SharedElementEncapsulation(object):
    """Encapsulates a reference to an element available in the shared module.

    This is used by Futures (map on lambda, for instance)."""
    def __init__(self, element):
        self.isMethod = False
        if utils.isStr(element):
            # Already shared element
            assert getConst(element, timeout=0) != None, (
                "Element must already be shared."
            )
            self.uniqueID = element
        else:
            # Element to share
            # Determine if function is a method. Methods derived from external
            # languages such as C++ aren't detected by ismethod.
            if ismethod(element):
                # Must share whole object before ability to use its method
                self.isMethod = True
                self.methodName = element.__name__
                element = element.__self__

            # Lambda-like or unshared code to share
            uniqueID = str(scoop.worker) + str(id(element)) + str(hash(element))
            self.uniqueID = uniqueID
            if getConst(uniqueID, timeout=0) == None:
                funcRef = {uniqueID: element}
                setConst(**funcRef)

    def __repr__(self):
        return self.uniqueID

    def __call__(self, *args, **kwargs):
        if self.isMethod:
            wholeObj = getConst(
                self.__repr__(),
                timeout=float("inf"),
            )
            return getattr(wholeObj, self.methodName)(*args, **kwargs)
        else:
            return getConst(self.__repr__(),
                            timeout=float("inf"))(*args, **kwargs)

    def __name__(self):
        return self.__repr__()



# -*- coding: utf-8 -*-
from Tree import *
import sys
import time


def main():
    executeTree()

importTree(sys.argv[1] if len(sys.argv) > 1 else "tree.txt")
registerMap(map)
    
if __name__=="__main__":
    bt = time.time()
    main()
    totalTime = time.time() - bt
    print("Serial total time : {}".format(totalTime))


#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
"""
Sums the mutliples of 3 and 5 below 1000000
"""
from time import time
from scoop import futures
from operator import add

def multiples(n):
    return set(range(0, 1000000, n))


if __name__ == '__main__':
    bt = time()
    serial_result = sum(set.union(*map(multiples, [3, 5])))
    serial_time = time() - bt

    bt = time()
    parallel_result = sum(futures.mapReduce(multiples, set.union, [3, 5]))
    parallel_reduce_time = time() - bt

    assert serial_result == parallel_result

    print("Serial time: {0:.4f} s\nParallel time: {1:.4f} s"
          "".format(serial_time, parallel_reduce_time)
    )



import argparse
import Tree

if __name__=="__main__":

    parser = argparse.ArgumentParser(description="Creates a random tree "
                                                 "structure")
    parser.add_argument('--height',
                          help='The height of the tree.',
                          type=int,
                          default=5)
    parser.add_argument('--minChildren', '-m',
                           help="The minimum number of children by nodes",
                           type=int,
                           default=0)
    parser.add_argument('--maxChildren', '-M',
                           help="The maximum number of children by nodes",
                           type=int,
                           default=8)
    parser.add_argument('--alpha', '-a',
                           help="The 'a' parameter of the Weibull distribution "
                           "used to create the task length.",
                           type=float,
                           default=1)
    parser.add_argument('--scale', '-s',
                           help="The mean time of each task.",
                           type = float,
                           default=1)
    parser.add_argument('--filename', '-f',
                           help="The filename to save the tree.",
                           default="tree.txt")

    args = parser.parse_args()
    
    if args.minChildren > args.maxChildren:
        args.maxChildren = args.minChildren + 1
    
    tree = Tree.Tree(args.height,
                args.minChildren,
                args.maxChildren,
                args.alpha,
                *Tree.calibrate(args.scale))
    Tree.exportTree(tree, args.filename)
    print("Generated :\n{0}".format(tree))



#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
import os
from inspect import ismethod
from collections import namedtuple, Iterable
from functools import partial, reduce
import itertools
import copy
import time

import scoop
from ._types import Future, CallbackType
from . import _control as control
from .fallbacks import (
    ensureScoopStartedProperlyMapFallback,
    ensureScoopStartedProperly,
    NotStartedProperly
)


# Constants stated by PEP 3148 (http://www.python.org/dev/peps/pep-3148/#module-functions)
FIRST_COMPLETED = 'FIRST_COMPLETED'
FIRST_EXCEPTION = 'FIRST_EXCEPTION'
ALL_COMPLETED = 'ALL_COMPLETED'
_AS_COMPLETED = '_AS_COMPLETED'

# This is the greenlet for running the controller logic.
_controller = None
callbackGroupID = itertools.count()


def _startup(rootFuture, *args, **kargs):
    """Initializes the SCOOP environment.

    :param rootFuture: Any callable object (function or class object with *__call__*
        method); this object will be called once and allows the use of parallel
        calls inside this object.
    :param args: A tuple of positional arguments that will be passed to the
        callable object.
    :param kargs: A dictionary of additional keyword arguments that will be
        passed to the callable object.

    :returns: The result of the root Future.

    Be sure to launch your root Future using this method."""
    import greenlet
    global _controller
    _controller = greenlet.greenlet(control.runController)
    try:
        result = _controller.switch(rootFuture, *args, **kargs)
    except scoop._comm.Shutdown:
        result = None
        control.execQueue.shutdown()
    if scoop.DEBUG:
        from scoop import _debug
        _debug.writeWorkerDebug(control.debug_stats,
                                control.QueueLength)
    return result


def _mapFuture(callable_, *iterables):
    """Similar to the built-in map function, but each of its
    iteration will spawn a separate independent parallel Future that will run
    either locally or remotely as `callable(*args)`.

    :param callable: Any callable object (function or class object with *__call__*
        method); this object will be called to execute each Future.
    :param iterables: A tuple of iterable objects; each will be zipped
        to form an iterable of arguments tuples that will be passed to the
        callable object as a separate Future.

    :returns: A list of Future objects, each corresponding to an iteration of
        map.

    On return, the Futures are pending execution locally, but may also be
    transfered remotely depending on global load. Execution may be carried on
    with any further computations. To retrieve the map results, you need to
    either wait for or join with the spawned Futures. See functions waitAny,
    waitAll, or joinAll. Alternatively, You may also use functions mapWait or
    mapJoin that will wait or join before returning."""
    childrenList = []
    for args in zip(*iterables):
        childrenList.append(submit(callable_, *args))
    return childrenList

def _mapGenerator(futures):
    """Generator function that iterates through the results in-order."""
    for future in _waitAll(*futures):
        yield future.resultValue


@ensureScoopStartedProperlyMapFallback
def map(func, *iterables, **kwargs):
    """map(func, *iterables)
    Equivalent to
    `map(func, \*iterables, ...)
    <http://docs.python.org/library/functions.html#map>`_
    but *func* is executed asynchronously
    and several calls to func may be made concurrently. This non-blocking call
    returns an iterator which raises a TimeoutError if *__next__()* is called
    and the result isn't available after timeout seconds from the original call
    to *map()*. If timeout is not specified or None then there is no limit to
    the wait time. If a call raises an exception then that exception will be
    raised when its value is retrieved from the iterator.

    :param func: Any picklable callable object (function or class object with
        *__call__* method); this object will be called to execute the Futures.
        The callable must return a value.
    :param iterables: Iterable objects; each will be zipped to form an iterable
        of arguments tuples that will be passed to the callable object as a
        separate Future.
    :param timeout: The maximum number of seconds to wait. If None, then there
        is no limit on the wait time.

    :returns: A generator of map results, each corresponding to one map
        iteration."""
    # TODO: Handle timeout
    futures = _mapFuture(func, *iterables)
    return _mapGenerator(futures)


def map_as_completed(func, *iterables, **kwargs):
    """map_as_completed(func, *iterables)
    Equivalent to map, but the results are returned as soon as they are made
    available.

    :param func: Any picklable callable object (function or class object with
        *__call__* method); this object will be called to execute the Futures.
        The callable must return a value.
    :param iterables: Iterable objects; each will be zipped to form an iterable
        of arguments tuples that will be passed to the callable object as a
        separate Future.
    :param timeout: The maximum number of seconds to wait. If None, then there
        is no limit on the wait time.

    :returns: A generator of map results, each corresponding to one map
        iteration."""
    # TODO: Handle timeout
    for future in as_completed(_mapFuture(func, *iterables)):
        yield future.resultValue


def _recursiveReduce(mapFunc, reductionFunc, scan, *iterables):
    """Generates the recursive reduction tree. Used by mapReduce."""
    if iterables:
        half = min(len(x) // 2 for x in iterables)
        data_left = [list(x)[:half] for x in iterables]
        data_right = [list(x)[half:] for x in iterables]
    else:
        data_left = data_right = [[]]

    # Submit the left and right parts of the reduction
    out_futures = [None, None]
    out_results = [None, None]
    for index, data in enumerate([data_left, data_right]):
        if any(len(x) <= 1 for x in data):
            out_results[index] = mapFunc(*list(zip(*data))[0])
        else:
            out_futures[index] = submit(
                _recursiveReduce,
                mapFunc,
                reductionFunc,
                scan,
                *data
            )

    # Wait for the results
    for index, future in enumerate(out_futures):
        if future:
            out_results[index] = future.result()

    # Apply a scan if needed
    if scan:
        last_results = copy.copy(out_results)
        if type(out_results[0]) is not list:
            out_results[0] = [out_results[0]]
        else:
            last_results[0] = out_results[0][-1]
        if type(out_results[1]) is list:
            out_results[0].extend(out_results[1][:-1])
            last_results[1] = out_results[1][-1]
        out_results[0].append(reductionFunc(*last_results))
        return out_results[0]

    return reductionFunc(*out_results)


@ensureScoopStartedProperly
def mapScan(mapFunc, reductionFunc, *iterables, **kwargs):
    """Exectues the :meth:`~scoop.futures.map` function and then applies a
    reduction function to its result while keeping intermediate reduction
    values. This is a blocking call.

    :param mapFunc: Any picklable callable object (function or class object with
        *__call__* method); this object will be called to execute the Futures.
        The callable must return a value.
    :param reductionFunc: Any picklable callable object (function or class object
        with *__call__* method); this object will be called to reduce pairs of
        Futures results. The callable must support two parameters and return a
        single value.
    :param iterables: Iterable objects; each will be zipped to form an iterable
        of arguments tuples that will be passed to the mapFunc object as a
        separate Future.
    :param timeout: The maximum number of seconds to wait. If None, then there
        is no limit on the wait time.

    :returns: Every return value of the reduction function applied to every
              mapped data sequentially ordered."""
    return submit(
        _recursiveReduce,
        mapFunc,
        reductionFunc,
        True,
        *iterables
    ).result()


@ensureScoopStartedProperly
def mapReduce(mapFunc, reductionFunc, *iterables, **kwargs):
    """Exectues the :meth:`~scoop.futures.map` function and then applies a
    reduction function to its result. The reduction function will cumulatively
    merge the results of the map function in order to get a single final value.
    This call is blocking.

    :param mapFunc: Any picklable callable object (function or class object
        with *__call__* method); this object will be called to execute the
        Futures. The callable must return a value.
    :param reductionFunc: Any picklable callable object (function or class
        object with *__call__* method); this object will be called to reduce
        pairs of Futures results. The callable must support two parameters and
        return a single value.
    :param iterables: Iterable objects; each will be zipped to form an iterable
        of arguments tuples that will be passed to the callable object as a
        separate Future.
    :param timeout: The maximum number of seconds to wait. If None, then there
        is no limit on the wait time.

    :returns: A single value."""
    return submit(
        _recursiveReduce,
        mapFunc,
        reductionFunc,
        False,
        *iterables
    ).result()


def _createFuture(func, *args):
    """Helper function to create a future."""
    assert callable(func), (
        "The provided func parameter is not a callable."
    )

    # If function is a lambda or class method, share it (or its parent object)
    # beforehand
    lambdaType = type(lambda: None)
    funcIsLambda = isinstance(func, lambdaType) and func.__name__ == '<lambda>'
    # Determine if function is a method. Methods derived from external
    # languages such as C++ aren't detected by ismethod.
    funcIsMethod = ismethod(func)
    if funcIsLambda or funcIsMethod:
        from .shared import SharedElementEncapsulation
        func = SharedElementEncapsulation(func)

    return Future(control.current.id, func, *args)

@ensureScoopStartedProperly
def submit(func, *args):
    """Submit an independent asynchronous :class:`~scoop._types.Future` that will
    either run locally or remotely as `func(*args)`.

    :param func: Any picklable callable object (function or class object with
        *__call__* method); this object will be called to execute the Future.
        The callable must return a value.
    :param args: A tuple of positional arguments that will be passed to the
        func object.

    :returns: A future object for retrieving the Future result.

    On return, the Future can be pending execution locally but may also be
    transfered remotely depending on load or on remote distributed workers. You
    may carry on with any further computations while the Future completes.
    Result retrieval is made via the :meth:`~scoop._types.Future.result`
    function on the Future."""
    child = _createFuture(func, *args)

    control.futureDict[control.current.id].children[child] = None
    control.execQueue.append(child)
    return child


def _waitAny(*children):
    """Waits on any child Future created by the calling Future.

    :param children: A tuple of children Future objects spawned by the calling
        Future.

    :return: A generator function that iterates on futures that are done.

    The generator produces results of the children in a non deterministic order
    that depends on the particular parallel execution of the Futures. The
    generator returns a tuple as soon as one becomes available."""
    n = len(children)
    # check for available results and index those unavailable
    for index, future in enumerate(children):
        if future.exceptionValue:
            raise future.exceptionValue
        if future._ended():
            future._delete()
            yield future
            n -= 1
        else:
            future.index = index
    future = control.current
    while n > 0:
        # wait for remaining results; switch to controller
        future.stopWatch.halt()
        childFuture = _controller.switch(future)
        future.stopWatch.resume()
        if childFuture.exceptionValue:
            raise childFuture.exceptionValue
        # Only yield if executed future was in children, otherwise loop
        if childFuture in children:
            childFuture._delete()
            yield childFuture
            n -= 1


def _waitAll(*children):
    """Wait on all child futures specified by a tuple of previously created
       Future.

    :param children: A tuple of children Future objects spawned by the calling
        Future.

    :return: A generator function that iterates on Future results.

    The generator produces results in the order that they are specified by
    the children argument. Because Futures are executed in a non deterministic
    order, the generator may have to wait for the last result to become
    available before it can produce an output. See waitAny for an alternative
    option."""
    for future in children:
        for f in _waitAny(future):
            yield f


def wait(fs, timeout=-1, return_when=ALL_COMPLETED):
    """Wait for the futures in the given sequence to complete.
    Using this function may prevent a worker from executing.

    :param fs: The sequence of Futures to wait upon.
    :param timeout: The maximum number of seconds to wait. If negative or not
        specified, then there is no limit on the wait time.
    :param return_when: Indicates when this function should return. The options
        are:

        ===============   ================================================
        FIRST_COMPLETED   Return when any future finishes or is cancelled.
        FIRST_EXCEPTION   Return when any future finishes by raising an
                          exception. If no future raises an exception then
                          it is equivalent to ALL_COMPLETED.
        ALL_COMPLETED     Return when all futures finish or are cancelled.
        ===============   ================================================

    :return: A named 2-tuple of sets. The first set, named 'done', contains the
        futures that completed (is finished or cancelled) before the wait
        completed. The second set, named 'not_done', contains uncompleted
        futures."""

    DoneAndNotDoneFutures = namedtuple('DoneAndNotDoneFutures', 'done not_done')
    if timeout < 0:
        # Negative timeout means blocking.
        if return_when == FIRST_COMPLETED:
            next(_waitAny(*fs))
        elif return_when in [ALL_COMPLETED, FIRST_EXCEPTION]:
            for _ in _waitAll(*fs):
                pass
        done = set(f for f in fs if f.done())
        not_done = set(fs) - done
        return DoneAndNotDoneFutures(done, not_done)

    elif timeout == 0:
        # Zero-value entry means non-blocking
        control.execQueue.flush()
        control.execQueue.updateQueue()
        done = set(f for f in fs if f._ended())
        not_done = set(fs) - done
        return DoneAndNotDoneFutures(done, not_done)

    else:
        # Any other value means blocking for a given time.
        done = set()
        start_time = time.time()
        while time.time() - start_time < timeout:
            # Flush futures on local queue (to be executed remotely)
            control.execQueue.flush()
            # Block until data arrives (to free CPU time)
            control.execQueue.socket._poll(time.time() - start_time)
            # Update queue
            control.execQueue.updateQueue()

            for f in fs:
                if f._ended():
                    done.add(f)

            not_done = set(fs) - done

            if return_when == FIRST_COMPLETED and len(done) > 0:
                break
            if len(not_done) == 0:
                break
        return DoneAndNotDoneFutures(done, not_done)


def as_completed(fs, timeout=None):
    """Iterates over the given futures that yields each as it completes. This
    call is blocking.

    :param fs: The sequence of Futures to wait upon.
    :param timeout: The maximum number of seconds to wait. If None, then there
        is no limit on the wait time.

    :return: An iterator that yields the given Futures as they complete
        (finished or cancelled).
    """
    #TODO: Handle timeout
    return _waitAny(*fs)


def _join(child):
    """This private function is for joining the current Future with one of its
    child Future.

    :param child: A child Future object spawned by the calling Future.

    :return: The result of the child Future.

    Only one Future can be specified. The function returns a single
    corresponding result as soon as it becomes available."""
    for future in _waitAny(child):
        return future.resultValue


def _joinAll(*children):
    """This private function is for joining the current Future with all of the
    children Futures specified in a tuple.

    :param children: A tuple of children Future objects spawned by the calling
        Future.

    :return: A list of corresponding results for the children Futures.

    This function will wait for the completion of all specified child Futures
    before returning to the caller."""
    return [_join(future) for future in _waitAll(*children)]


def shutdown(wait=True):
    """This function is here for compatibility with `futures` (PEP 3148) and
    doesn't have any behavior.

    :param wait: Unapplied parameter."""
    pass



#!/usr/bin/env python
#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
import scoop
scoop.DEBUG = False

import unittest
import subprocess
import time
import copy
import os
import sys
import operator
import signal
import math
from tests_parser import TestUtils
from tests_stat import TestStat
from tests_stopwatch import TestStopWatch

from scoop import futures, _control, utils, shared
from scoop._types import FutureQueue
from scoop.broker.structs import BrokerInfo


subprocesses = []
def cleanSubprocesses():
    [a.kill() for a in subprocesses]

try:
    signal.signal(signal.SIGQUIT, cleanSubprocesses)
except AttributeError:
    # SIGQUIT doesn't exist on Windows
    signal.signal(signal.SIGTERM, cleanSubprocesses)


def func0(n):
    task = futures.submit(func1, n)
    result = task.result()
    return result


def func1(n):
    result = futures.map(func2, [i+1 for i in range(n)])
    return sum(result)


def func2(n):
    launches = []
    for i in range(n):
        launches.append(futures.submit(func3, i + 1))
    result = futures.as_completed(launches)
    return sum(r.result() for r in result)


def func3(n):
    result = list(futures.map(func4, [i+1 for i in range(n)]))
    return sum(result)


def func4(n):
    result = n * n
    return result


def funcLambda(n):
    lambda_func = lambda x : x*x
    result = list(futures.map(lambda_func, [i+1 for i in range(n)]))
    return sum(result)


def funcLambdaSubfuncNotGlobal(n):
    """Tests a lambda function containing a call to a function that is not in
    the globals()."""
    my_mul = operator.mul
    lambda_func = lambda x : my_mul(x, x)
    result = list(futures.map(lambda_func, [i+1 for i in range(n)]))
    return sum(result)


def funcCos():
    result = list(futures.map(math.cos, [i for i in range(10)]))
    return sum(result)


def funcCallback():
    f = futures.submit(func4, 100)

    def callBack(future):
        future.was_callabacked = True

    f.add_done_callback(callBack)
    if len(f.callback) == 0:
        return False
    futures.wait((f,))
    try:
        return f.was_callabacked
    except:
        return False


def funcCancel():
    f = futures.submit(func4, 100)
    f.cancel()
    return f.cancelled()


def funcCompleted(n):
    launches = []
    for i in range(n):
        launches.append(futures.submit(func4, i + 1))
    result = futures.as_completed(launches)
    return sum(r.result() for r in result)


def funcDone():
    f = futures.submit(func4, 100)
    futures.wait((f,))
    done = f.done()
    if done != True:
        return done
    res = f.result()
    done = f.done()
    return done

def funcWait(timeout):
    fs = [futures.submit(func4, i) for i in range(1000)]
    done, not_done = futures.wait(fs, timeout=timeout)
    return done, not_done


def funcExcept(n):
    f = futures.submit(funcRaise, n)
    try:
        f.result()
    except:
        return True

    return False


def funcRaise(n):
    raise Exception("Test exception")


def funcSub(n):
    f = futures.submit(func4, n)
    return f.result()


def funcMapScan(l):
    resultat = futures.mapScan(func4,
                               operator.add,
                               l)
    time.sleep(0.3)
    _control.execQueue.socket.pumpInfoSocket()
    return resultat


def funcMapReduce(l):
    resultat = futures.mapReduce(func4,
                                 operator.add,
                                 l)
    time.sleep(0.3)
    _control.execQueue.socket.pumpInfoSocket()
    return resultat

def funcDoubleMapReduce(l):
    resultat = futures.mapReduce(func4,
                                 operator.add,
                                 l)
    resultat2 = futures.mapReduce(func4,
                                 operator.add,
                                 l)
    time.sleep(0.3)
    _control.execQueue.socket.pumpInfoSocket()
    return resultat == resultat2


def funcUseSharedConstant():
    # Tries on a mutable and an immutable object
    assert shared.getConst('myVar') == {
        1: 'Example 1',
        2: 'Example 2',
        3: 'Example 3',
    }
    assert shared.getConst('secondVar') == "Hello World!"
    return True


def funcUseSharedFunction():
    assert shared.getConst('myRemoteFunc')(5) == 5 * 5
    assert shared.getConst('myRemoteFunc')(25) == 25 * 25
    return True


def funcSharedConstant():
    shared.setConst(myVar={1: 'Example 1',
                                2: 'Example 2',
                                3: 'Example 3',
                               })
    shared.setConst(secondVar="Hello World!")
    result = True
    for _ in range(100):
        try:
            result &= futures.submit(funcUseSharedConstant).result()
        except AssertionError:
            result = False
    return result


def funcSharedFunction():
    shared.setConst(myRemoteFunc=func4)
    result = True
    for _ in range(100):
        try:
            result &= futures.submit(funcUseSharedFunction).result()
        except AssertionError:
            result = False
    return result


def funcMapAsCompleted(n):
    result = list(futures.map_as_completed(func4, [i+1 for i in range(n)]))
    return sum(result)


def funcIter(n):
    result = list(futures.map(func4, (i+1 for i in range(n))))
    return sum(result)


def main(n):
    task = futures.submit(func0, n)
    futures.wait([task], return_when=futures.ALL_COMPLETED)
    result = task.result()
    return result


def main_simple(n):
    task = futures.submit(func3, n)
    futures.wait([task], return_when=futures.ALL_COMPLETED)
    result = task.result()
    return result


def submit_get_queues_size(n):
    task = futures.submit(func4, n)
    result = task.result()
    return [
        len(scoop._control.execQueue.inprogress),
        len(scoop._control.execQueue.ready),
        len(scoop._control.execQueue.movable),
        len(scoop._control.futureDict) - 1, # - 1 because the current function is a future too
    ]


def map_get_queues_size(n):
    result = list(map(func4, [n for n in range(n)]))
    return [
        len(scoop._control.execQueue.inprogress),
        len(scoop._control.execQueue.ready),
        len(scoop._control.execQueue.movable),
        len(scoop._control.futureDict) - 1, # - 1 because the current function is a future too
    ]


def port_ready(port, socket):
    """Checks if a given port is already binded"""
    try:
        socket.connect(('127.0.0.1', port))
    except IOError:
        return False
    else:
        socket.shutdown(2)
        socket.close()
        return True


class TestScoopCommon(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        # Parent initialization
        super(TestScoopCommon, self).__init__(*args, **kwargs)

    def multiworker_set(self):
        global subprocesses
        worker = subprocess.Popen([sys.executable, "-m", "scoop.bootstrap",
        "--brokerHostname", "127.0.0.1", "--taskPort", "5555",
        "--metaPort", "5556", "tests.py"])
        subprocesses.append(worker)
        return worker

    def setUp(self):
        global subprocesses
        # Start the server
        self.server = subprocess.Popen([sys.executable, "-m", "scoop.broker",
        "--tPort", "5555", "--mPort", "5556"])
        import socket, datetime, time
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        begin = datetime.datetime.now()
        while not port_ready(5555, s):
            if (datetime.datetime.now() - begin > datetime.timedelta(seconds=3)):
                raise Exception('Could not start server!')
            pass
        subprocesses.append(self.server)
        scoop.IS_RUNNING = True
        scoop.IS_ORIGIN = True
        scoop.WORKER_NAME = 'origin'.encode()
        scoop.BROKER_NAME = 'broker'.encode()
        scoop.BROKER = BrokerInfo("127.0.0.1",
                                  5555,
                                  5556,
                                  "127.0.0.1")
        scoop.worker = (scoop.WORKER_NAME, scoop.BROKER_NAME)
        scoop.MAIN_MODULE = "tests.py"
        scoop.VALID = True
        scoop.DEBUG = False
        scoop.SIZE = 2
        scoop._control.execQueue = FutureQueue()

    def tearDown(self):
        global subprocesses
        scoop._control.futureDict.clear()
        try: self.w.kill()
        except: pass
        # Destroy the server
        if self.server.poll() == None:
            try: self.server.kill()
            except: pass
        # Stabilise zmq after a deleted socket
        del subprocesses[:]
        time.sleep(0.1)


class TestMultiFunction(TestScoopCommon):
    def __init__(self, *args, **kwargs):
        # Parent initialization
        super(TestMultiFunction, self).__init__(*args, **kwargs)
        self.main_func = main
        self.small_result = 77
        self.large_result = 76153

    def test_small_uniworker(self):
        _control.FutureQueue.highwatermark = 10
        _control.FutureQueue.lowwatermark = 5
        result = futures._startup(self.main_func, 4)
        self.assertEqual(result, self.small_result)

    def test_small_no_lowwatermark_uniworker(self):
        _control.FutureQueue.highwatermark = 9999999999999
        _control.FutureQueue.lowwatermark = 1
        result = futures._startup(self.main_func, 4)
        self.assertEqual(result, self.small_result)

    def test_small_foreign_uniworker(self):
        _control.FutureQueue.highwatermark = 1
        result = futures._startup(self.main_func, 4)
        self.assertEqual(result, self.small_result)

    def test_small_local_uniworker(self):
        _control.FutureQueue.highwatermark = 9999999999999
        result = futures._startup(self.main_func, 4)
        self.assertEqual(result, self.small_result)

    def test_large_uniworker(self):
        _control.FutureQueue.highwatermark = 9999999999999
        result = futures._startup(self.main_func, 20)
        self.assertEqual(result, self.large_result)

    def test_large_no_lowwatermark_uniworker(self):
        _control.FutureQueue.lowwatermark = 1
        _control.FutureQueue.highwatermark = 9999999999999
        result = futures._startup(self.main_func, 20)
        self.assertEqual(result, self.large_result)

    def test_large_foreign_uniworker(self):
        _control.FutureQueue.highwatermark = 1
        result = futures._startup(self.main_func, 20)
        self.assertEqual(result, self.large_result)

    def test_large_local_uniworker(self):
        _control.FutureQueue.highwatermark = 9999999999999
        result = futures._startup(self.main_func, 20)
        self.assertEqual(result, self.large_result)

    def test_small_local_multiworker(self):
        self.w = self.multiworker_set()
        _control.FutureQueue.highwatermark = 9999999999999
        Backupenv = os.environ.copy()
        result = futures._startup(self.main_func, 4)
        self.assertEqual(result, self.small_result)
        os.environ = Backupenv

    def test_small_foreign_multiworker(self):
        self.w = self.multiworker_set()
        _control.FutureQueue.highwatermark = 1
        Backupenv = os.environ.copy()
        result = futures._startup(self.main_func, 4)
        self.assertEqual(result, self.small_result)
        os.environ = Backupenv

    def test_execQueue_multiworker(self):
        self.w = self.multiworker_set()
        result = futures._startup(func0, 6)
        self.assertEqual(len(scoop._control.execQueue.inprogress), 0)
        self.assertEqual(len(scoop._control.execQueue.ready), 0)
        self.assertEqual(len(scoop._control.execQueue.movable), 0)
        self.assertEqual(len(scoop._control.futureDict), 0)

    def test_execQueue_uniworker(self):
        result = futures._startup(func0, 6)
        self.assertEqual(len(scoop._control.execQueue.inprogress), 0)
        self.assertEqual(len(scoop._control.execQueue.ready), 0)
        self.assertEqual(len(scoop._control.execQueue.movable), 0)
        self.assertEqual(len(scoop._control.futureDict), 0)

    def test_execQueue_submit_uniworker(self):
        result = futures._startup(submit_get_queues_size, 6)
        self.assertEqual(
            result,
            [0 for _ in range(len(result))],
            "Buffers are not empty after future completion"
        )

    def test_execQueue_map_uniworker(self):
        result = futures._startup(map_get_queues_size, 6)
        self.assertEqual(
            result,
            [0 for _ in range(len(result))],
            "Buffers are not empty after future completion"
        )

    def test_execQueue_submit_multiworker(self):
        self.w = self.multiworker_set()
        result = futures._startup(submit_get_queues_size, 6)
        self.assertEqual(
            result,
            [0 for _ in range(len(result))],
            "Buffers are not empty after future completion"
        )

    def test_execQueue_map_multiworker(self):
        self.w = self.multiworker_set()
        result = futures._startup(map_get_queues_size, 6)
        self.assertEqual(
            result,
            [0 for _ in range(len(result))],
            "Buffers are not empty after future completion"
        )


    def test_partial(self):
        """This function removes some attributes (such as __name__)."""
        from functools import partial
        result = futures._startup(partial(self.main_func), 4)
        self.assertEqual(result, self.small_result)


class TestSingleFunction(TestMultiFunction):
    def __init__(self, *args, **kwargs):
        # Parent initialization
        super(TestSingleFunction, self).__init__(*args, **kwargs)
        self.main_func = main_simple
        self.small_result = 30
        self.large_result = 2870


class TestApi(TestScoopCommon):
    def __init(self, *args, **kwargs):
        super(TestApi, self).__init(*args, **kwargs)

    def test_as_Completed_single(self):
        result = futures._startup(funcCompleted, 30)
        self.assertEqual(result, 9455)

    def test_as_Completed_multi(self):
        self.w = self.multiworker_set()
        result = futures._startup(funcCompleted, 30)
        self.assertEqual(result, 9455)

    def test_map_single(self):
        result = futures._startup(func3, 30)
        self.assertEqual(result, 9455)

    def test_map_multi(self):
        self.w = self.multiworker_set()
        result = futures._startup(func3, 30)
        self.assertEqual(result, 9455)

    def test_map_lambda(self):
        self.w = self.multiworker_set()
        result = futures._startup(funcLambda, 30)
        self.assertEqual(result, 9455)

    # This test is complex to handle and has many implications
    # Bundle a closure with the future?
    # How to manage side-effects of variables in closure?
    #def test_map_lambda_subfunc_not_global(self):
    #    self.w = self.multiworker_set()
    #    result = futures._startup(funcLambdaSubfuncNotGlobal, 30)
    #    self.assertEqual(result, 9455)

    def test_map_imported_func(self):
        self.w = self.multiworker_set()
        result = futures._startup(funcCos)
        self.assertGreater(result, 0.4)
        self.assertLess(result, 0.5)

    def test_submit_single(self):
        result = futures._startup(funcSub, 10)
        self.assertEqual(result, 100)

    def test_submit_multi(self):
        self.w = self.multiworker_set()
        result = futures._startup(funcSub, 10)
        self.assertEqual(result, 100)

    def test_exception_single(self):
        result = futures._startup(funcExcept, 19)
        self.assertTrue(result)

    def test_exception_multi(self):
        self.w = self.multiworker_set()
        result = futures._startup(funcExcept, 19)
        self.assertTrue(result)

    def test_done(self):
        result = futures._startup(funcDone)
        self.assertTrue(result)

    def test_cancel(self):
        self.assertTrue(futures._startup(funcCancel))

    def test_callback(self):
        self.assertTrue(futures._startup(funcCallback))

    def test_wait_no_timeout(self):
        done, not_done = futures._startup(funcWait, -1)
        self.assertTrue(len(done) == 1000)
        self.assertTrue(len(not_done) == 0)

    def test_wait_with_timeout(self):
        done, not_done = futures._startup(funcWait, 0.1)
        self.assertTrue((len(done) + len(not_done)) == 1000)

    def test_wait_nonblocking(self):
        done, not_done = futures._startup(funcWait, 0)
        self.assertTrue((len(done) + len(not_done)) == 1000)

    def test_map_as_completed_single(self):
        result = futures._startup(funcMapAsCompleted, 30)
        self.assertEqual(result, 9455)

    def test_map_as_completed_multi(self):
        self.w = self.multiworker_set()
        result = futures._startup(funcMapAsCompleted, 30)
        self.assertEqual(result, 9455)

    def test_from_generator_single(self):
        result = futures._startup(funcIter, 30)
        self.assertEqual(result, 9455)

    def test_from_generator_multi(self):
        self.w = self.multiworker_set()
        result = futures._startup(funcIter, 30)
        self.assertEqual(result, 9455)


class TestCoherent(TestScoopCommon):
    def __init(self, *args, **kwargs):
        super(TestCoherent, self).__init(*args, **kwargs)

    def test_mapReduce(self):
        result = futures._startup(funcMapReduce, [10, 20, 30])
        self.assertEqual(result, 1400)

    def test_doubleMapReduce(self):
        result = futures._startup(funcDoubleMapReduce, [10, 20, 30])
        self.assertTrue(result)

    def test_mapScan(self):
        result = futures._startup(funcMapScan, [10, 20, 30])
        self.assertEqual(max(result), 1400)


class TestShared(TestScoopCommon):
    def __init(self, *args, **kwargs):
        super(TestShared, self).__init(*args, **kwargs)

    def test_shareConstant(self):
        result = futures._startup(funcSharedFunction)
        self.assertEqual(result, True)

    def test_shareFunction(self):
        result = futures._startup(funcSharedConstant)
        self.assertEqual(result, True)


if __name__ == '__main__' and os.environ.get('IS_ORIGIN', "1") == "1":
    utSimple = unittest.TestLoader().loadTestsFromTestCase(TestSingleFunction)
    utComplex = unittest.TestLoader().loadTestsFromTestCase(TestMultiFunction)
    utApi = unittest.TestLoader().loadTestsFromTestCase(TestApi)
    utUtils = unittest.TestLoader().loadTestsFromTestCase(TestUtils)
    utCoherent = unittest.TestLoader().loadTestsFromTestCase(TestCoherent)
    utShared = unittest.TestLoader().loadTestsFromTestCase(TestShared)
    utStat = unittest.TestLoader().loadTestsFromTestCase(TestStat)
    utStopWatch = unittest.TestLoader().loadTestsFromTestCase(TestStopWatch)

    if len(sys.argv) > 1:
        if sys.argv[1] == "simple":
            unittest.TextTestRunner(verbosity=2).run(utSimple)
        elif sys.argv[1] == "complex":
            unittest.TextTestRunner(verbosity=2).run(utComplex)
        elif sys.argv[1] == "api":
            unittest.TextTestRunner(verbosity=2).run(utApi)
        elif sys.argv[1] == "utils":
            unittest.TextTestRunner(verbosity=2).run(utUtils)
        elif sys.argv[1] == "coherent":
            unittest.TextTestRunner(verbosity=2).run(utCoherent)
        elif sys.argv[1] == "shared":
            unittest.TextTestRunner(verbosity=2).run(utShared)
        elif sys.argv[1] == "stat":
            unittest.TextTestRunner(verbosity=2).run(utStat)
        elif sys.argv[1] == "stopwatch":
            unittest.TextTestRunner(verbosity=2).run(utStopWatch)
        elif sys.argv[1] == "verbose":
            sys.argv = sys.argv[0:1]
            unittest.main(verbosity=2)
    else:
        unittest.main()
elif __name__ == '__main__':
    futures._startup(main_simple)



#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
from . import minusconf
import scoop

SERVICES_DISCOVERED = []


class SCOOPool(object):
    def __init__(self, host, ports, name):
        self._host = host
        self._ports = ports
        self._name = name

    @property
    def host(self):
        """Get address of discovered broker.
        Will convert back to IPv4 accordingly."""
        # TODO Is %interfacename working?
        host = self._host
        if host[:7] == "::ffff:":
            # IPv4 address mapped to IPv6, convert back
            host = host[7:]
        return host

    @property
    def ports(self):
        return self._ports.split(",")

    @property
    def name(self):
        return self._name


def _print_error(seeker, opposite, error_str):
    import sys
    sys.stderr.write("Error from {opposite}: {error_str}\n".format(
        opposite=opposite,
        error_str=error_str,
    ))


def _seekerCallback(seeker, svca):
    global SERVICES_DISCOVERED
    SERVICES_DISCOVERED.append(
        SCOOPool(
            svca.addr,
            svca.port,
            svca.aname,
        )
    )
    scoop.logger.info("Discovery seeker has found a broker.")


def Advertise(port, stype="SCOOP", sname="Broker", advertisername="Broker",
              location=""):
    """
    stype = always SCOOP
    port = comma separated ports
    sname = broker unique name
    location = routable location (ip or dns)
    """
    scoop.logger.info("Launching advertiser...")
    service = minusconf.Service(stype, port, sname, location)
    advertiser = minusconf.ThreadAdvertiser([service], advertisername)
    advertiser.start()
    scoop.logger.info("Advertiser launched.")

    return advertiser


def Seek(stype="SCOOP", sname="Broker", advertisername=""):
    scoop.logger.info("Launching discovery seeker...")
    se = minusconf.Seeker(stype=stype, aname=advertisername, sname=sname,
                          find_callback=_seekerCallback,
                          error_callback=_print_error,
                          )
    se.run()
    # TODO: Remove the timeout and return as soon as something is found.
    return SERVICES_DISCOVERED



from scoop import utils

import unittest
import os

# This is a subset of a SGE environment for testing
SGE_ENV = {'NHOSTS':'2', 'PE_HOSTFILE':"sgehostssim.txt",
'NSLOTS': '16', 'PE': 'default', 'ENVIRONMENT': 'BATCH'}

# This is a subset of a PBS environment for testing
PBS_ENV = {'ENVIRONMENT': 'BATCH', 'PBS_ENVIRONMENT': 'PBS_BATCH',
'MOAB_PROCCOUNT': '16', 'PBS_NUM_NODES': '2', 'PBS_NP': '16', 
'PBS_NODEFILE': 'pbshostssim.txt'} 

# This is the logical content of the hostfiles
hosts = [("host1", 8), ("host2", 4), ("host3", 2), ("host4", 2)]

class TestUtils(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestUtils, self).__init__(*args, **kwargs)

    def setUp(self):
        self.backup = os.environ.copy()
        

    def tearDown(self):
        os.environ = self.backup

    def test_getCpuCount(self):
        # Since we don't know how many cores has the computer on which these
        # tests will be run, we check that at least, getCPUcount() returns an
        # int larger than 0.
        self.assertIsInstance(utils.getCPUcount(), int)
        self.assertTrue(utils.getCPUcount() > 0)

    def test_getEnvPBS(self):
        os.environ.update(PBS_ENV)
        self.assertEqual(utils.getEnv(), "PBS")

    def test_getEnvSGE(self):
        os.environ.update(SGE_ENV)
        self.assertEqual(utils.getEnv(), "SGE")

    def test_getEnvOther(self):
        self.assertEqual(utils.getEnv(), "other")

    def test_getHostsPBS(self):
        os.environ.update(PBS_ENV)
        # We used the set because the order of the hosts is not important.
        self.assertEqual(set(utils.getHosts()), set(hosts))

    def test_getHostsSGE(self):
        os.environ.update(SGE_ENV)
        # We used the set because the order of the hosts is not important.
        self.assertEqual(set(utils.getHosts()), set(hosts))

    def test_getHostsFile(self):
        self.assertEqual(set(utils.getHosts("hostfilesim.txt")), set(hosts))

    def test_getWorkerQtePBS(self):
        os.environ.update(PBS_ENV)
        self.assertEqual(utils.getWorkerQte(utils.getHosts()), 16)

    def test_getWorkerQteSGE(self):
        os.environ.update(SGE_ENV)
        self.assertEqual(utils.getWorkerQte(utils.getHosts()), 16)

    def test_getWorkerQteFile(self):
        self.assertEqual(utils.getWorkerQte(utils.getHosts("hostfilesim.txt")), 16)





if __name__ == "__main__":
    t = unittest.TestLoader().loadTestsFromTestCase(TestUtils)
    unittest.TextTestRunner(verbosity=2).run(t)



#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
from collections import deque, defaultdict
import os
import time
import tempfile
import sys
import math
from functools import partial

import greenlet

from ._types import Future, FutureId, FutureQueue, CallbackType
import scoop

# Backporting collection features
if sys.version_info < (2, 7):
    from scoop.backports.newCollections import deque

# Future currently running in greenlet
current = None
# Dictionary of existing futures
futureDict = {}
# Queue of futures pending execution
execQueue = None
# Execution Statistics
class _stat(deque):
    def __init__(self, *args, **kargs):
        self._sum = 0
        self._squared_sum = 0
        super(_stat, self).__init__(*args, maxlen=10, **kargs)

    def appendleft(self, x):
        if len(self) >= self.maxlen:
            self._sum -= self[-1]
            self._squared_sum -= self[-1] ** 2
        self._sum += x
        self._squared_sum += x ** 2
        super(_stat, self).appendleft(x)

    def mean(self):
        ourSize = len(self)
        if ourSize > 3:
            return self._sum / ourSize
        return float("inf")

    def std(self):
        ourSize = len(self)
        if ourSize > 3:
            return math.sqrt(len(self) * self._squared_sum - self._sum ** 2) / len(self)
        return float("inf")


def advertiseBrokerWorkerDown(exctype, value, traceback):
    """Hook advertizing the broker if an impromptu shutdown is occuring."""
    if not scoop.SHUTDOWN_REQUESTED:
        execQueue.socket.shutdown()
    sys.__excepthook__(exctype, value, traceback)


def init_debug():
    """Initialise debug_stats and QueueLength (this is not a reset)"""
    global debug_stats
    global QueueLength
    if debug_stats is None:
        list_defaultdict = partial(defaultdict, list)
        debug_stats = defaultdict(list_defaultdict)
        QueueLength = []


def delFutureById(futureId, parentId):
    """Delete future on id basis"""
    try:
        del futureDict[futureId]
    except KeyError:
        pass
    try:
        toDel = [a for a in futureDict[parentId].children if a.id == futureId]
        for f in toDel:
            del futureDict[parentId].children[f]
    except KeyError:
        pass


def delFuture(afuture):
    """Delete future afuture"""
    try:
        del futureDict[afuture.id]
    except KeyError:
        pass
    try:
        del futureDict[afuture.parentId].children[afuture]
    except KeyError:
        pass


def runFuture(future):
    """Callable greenlet in charge of running tasks."""
    global debug_stats
    global QueueLength
    if scoop.DEBUG:
        init_debug()  # in case _control is imported before scoop.DEBUG was set
        debug_stats[future.id]['start_time'].append(time.time())
    future.waitTime = future.stopWatch.get()
    future.stopWatch.reset()
    # Get callback Group ID and assign the broker-wide unique executor ID
    try:
        uniqueReference = [cb.groupID for cb in future.callback][0]
    except IndexError:
        uniqueReference = None
    future.executor = (scoop.worker, uniqueReference)
    try:
        future.resultValue = future.callable(*future.args, **future.kargs)
    except BaseException as err:
        import traceback
        scoop.logger.debug(
            "The following error occurend on a worker:\n{err}\n{tb}".format(
                err=err,
                tb=traceback.format_exc(),
            )
        )
        future.exceptionValue = err
    future.executionTime = future.stopWatch.get()
    future.isDone = True

    # Update the worker inner work statistics
    if future.executionTime != 0. and hasattr(future.callable, '__name__'):
        execStats[future.callable.__name__].appendleft(future.executionTime)

    # Set debugging informations if needed
    if scoop.DEBUG:
        t = time.time()
        debug_stats[future.id]['end_time'].append(t)
        debug_stats[future.id].update({
            'executionTime': future.executionTime,
            'worker': scoop.worker,
            'creationTime': future.creationTime,
            'callable': str(future.callable.__name__)
                if hasattr(future.callable, '__name__')
                else 'No name',
           'parent': future.parentId
        })
        QueueLength.append((t, len(execQueue), execQueue.timelen(execQueue)))

    # Run callback (see http://www.python.org/dev/peps/pep-3148/#future-objects)
    future._execute_callbacks(CallbackType.universal)

    # Delete references to the future
    future._delete()

    return future


def runController(callable_, *args, **kargs):
    """Callable greenlet implementing controller logic."""
    global execQueue
    # initialize and run root future
    rootId = FutureId(-1, 0)

    # initialise queue
    if execQueue is None:
        execQueue = FutureQueue()

        sys.excepthook = advertiseBrokerWorkerDown

        # TODO: Make that a function
        # Wait until we received the main module if we are a headless slave
        headless = scoop.CONFIGURATION.get("headless", False)
        if not scoop.MAIN_MODULE:
            # If we're not the origin and still don't have our main_module,
            # wait for it and then import it as module __main___
            main = scoop.shared.getConst('__MAIN_MODULE__', timeout=float('inf'))
            directory_name = tempfile.mkdtemp()
            os.chdir(directory_name)
            scoop.MAIN_MODULE = main.writeFile(directory_name)
            from .bootstrap.__main__ import Bootstrap as SCOOPBootstrap
            newModule = SCOOPBootstrap.setupEnvironment()
            sys.modules['__main__'] = newModule
        elif scoop.IS_ORIGIN and headless and scoop.MAIN_MODULE:
            # We're the origin, share our main_module
            scoop.shared.setConst(
                __MAIN_MODULE__=scoop.encapsulation.ExternalEncapsulation(
                    scoop.MAIN_MODULE,
                )
            )
            # TODO: use modulefinder to share every local dependency of
            # main module

    # launch future if origin or try to pickup a future if slave worker
    if scoop.IS_ORIGIN:
        future = Future(rootId, callable_, *args, **kargs)
    else:
        future = execQueue.pop()

    future.greenlet = greenlet.greenlet(runFuture)
    future = future._switch(future)

    if scoop.DEBUG:
        lastDebugTs = time.time()

    while not scoop.IS_ORIGIN or future.parentId != rootId or not future._ended():
        if scoop.DEBUG and time.time() - lastDebugTs < scoop.TIME_BETWEEN_PARTIALDEBUG:
            from scoop import _debug
            _debug.writeWorkerDebug(
                debug_stats,
                QueueLength,
                "debug/partial-{0}".format(
                    round(time.time(), -1)
                )
            )
        # process future
        if future._ended():
            # future is finished
            if future.id.worker != scoop.worker:
                # future is not local
                execQueue.sendResult(future)
                future = execQueue.pop()
            else:
                # future is local, parent is waiting
                if future.index is not None:
                    parent = futureDict[future.parentId]
                    if parent.exceptionValue is None:
                        future = parent._switch(future)
                    else:
                        future = execQueue.pop()
                else:
                    future = execQueue.pop()
        else:
            # future is in progress; run next future from pending execution queue.
            future = execQueue.pop()

        if not future._ended() and future.greenlet is None:
            # initialize if the future hasn't started
            future.greenlet = greenlet.greenlet(runFuture)
            future = future._switch(future)

    execQueue.shutdown()
    if future.exceptionValue:
        raise future.exceptionValue
    return future.resultValue


execStats = defaultdict(_stat)

debug_stats = None
QueueLength = None
if scoop.DEBUG:
    init_debug()



#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
"""
Example of shared constants use.
This syntetic example only showcase the shared module API.
"""
from scoop import futures, shared
# Import SCOOP to get access to its constants such as scoop.worker
import scoop


def myFunc(parameter):
    """This function will be executed on the remote host even if it was not
       available at launch."""
    print('Hello World from {0}!'.format(scoop.worker))

    # It is possible to get a constant anywhere
    print(shared.getConst('myVar')[2])

    # Parameters are handled as usual
    return parameter + 1


if __name__ == "__main__":
    # Populate the shared constants
    shared.setConst(myVar={
        1: 'First element',
        2: 'Second element',
        3: 'Third element',
    })
    shared.setConst(secondVar="Hello World!")
    shared.setConst(myFunc=myFunc)

    # Use the previously defined shared function
    print(list(futures.map(myFunc, range(10))))
    
    # Un-commenting the following line will give a TypeError
    # since re-defining a constant is not allowed.
    #shared.setConst(myVar="Variable re-assignation")



#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
import random

from deap import algorithms
from deap import base
from deap import creator
from deap import tools
from scoop import futures

from dependency import sortingnetwork as sn
import logging
import time
import argparse
import sys

parser = argparse.ArgumentParser(description="Deap's evosn example.")
parser.add_argument('--inputs', type=int, default=6)
parser.add_argument('--cores', type=int, default=1)
parser.add_argument('--filename')
parser.add_argument('--population', type=int, default=300)
parser.add_argument('--generations', type=int, default=40)

args = parser.parse_args()
INPUTS = args.inputs

sizes = {11 : (29,39), 12 : (25,45), 13 : (39,51), 14 : (45,56),
         15 : (51,60), 16 : (56,65), 17 : (60,69), 18 : (65,74),
         19 : (69,78)}

def evalEvoSN(individual, dimension):
    network = sn.SortingNetwork(dimension, individual)
    return network.assess(), network.length, network.depth

def genWire(dimension):
    return (random.randrange(dimension), random.randrange(dimension))
    
def genNetwork(dimension, min_size, max_size):
    size = random.randint(min_size, max_size)
    return [genWire(dimension) for i in range(size)]
    
def mutWire(individual, dimension, indpb):
    for index, elem in enumerate(individual):
        if random.random() < indpb:
            individual[index] = genWire(dimension)      

def mutAddWire(individual, dimension):
    index = random.randint(0, len(individual))
    individual.insert(index, genWire(dimension))

def mutDelWire(individual):
    index = random.randrange(len(individual))
    del individual[index]

creator.create("FitnessMin", base.Fitness, weights=(-1.0, -1.0, -1.0))
creator.create("Individual", list, fitness=creator.FitnessMin)

toolbox = base.Toolbox()

# Gene initializer
toolbox.register("network", genNetwork, dimension=INPUTS,
        min_size=sizes[INPUTS][0] if INPUTS in sizes else 25,
        max_size=sizes[INPUTS][1] if INPUTS in sizes else 35)

# Structure initializers
toolbox.register("individual", tools.initIterate, creator.Individual, toolbox.network)
toolbox.register("population", tools.initRepeat, list, toolbox.individual)

toolbox.register("evaluate", evalEvoSN, dimension=INPUTS)
toolbox.register("mate", tools.cxTwoPoints)
toolbox.register("mutate", mutWire, dimension=INPUTS, indpb=0.05)
toolbox.register("addwire", mutAddWire, dimension=INPUTS)
toolbox.register("delwire", mutDelWire)
toolbox.register("select", tools.selTournament, tournsize=3)
toolbox.register("map", futures.map)
#logging.warning("avant main")
def main():
    # test if file is ok before starting the test
    if args.filename:
        open(args.filename).close()
    random.seed(64)
    
    beginTime = time.time()
    evaluationTime = 0

    population = toolbox.population(n=args.population)
    hof = tools.ParetoFront()
    
    stats = tools.Statistics(lambda ind: ind.fitness.values)
    stats.register("avg", tools.mean)
    stats.register("std", tools.std)
    stats.register("min", min)
    stats.register("max", max)
    
    logger = tools.EvolutionLogger(["gen", "evals", "time"] + [str(k) for k in
        stats.functions.keys()])
    logger.logHeader()

    CXPB, MUTPB, ADDPB, DELPB, NGEN = 0.5, 0.2, 0.01, 0.01, args.generations
    
    evalBegin = time.time()
    # Evaluate every individuals
    fitnesses = toolbox.map(toolbox.evaluate, population)

    
    for ind, fit in zip(population, fitnesses):
        ind.fitness.values = fit

    evaluationTime += (time.time() - evalBegin)
    
    hof.update(population)
    stats.update(population)
    
    logger.logGeneration(gen=0, evals=len(population), stats=stats, time=evaluationTime)
    
    # Begin the evolution
    for g in range(1, NGEN):
        offspring = [toolbox.clone(ind) for ind in population]
    
        # Apply crossover and mutation
        for ind1, ind2 in zip(offspring[::2], offspring[1::2]):
            if random.random() < CXPB:
                toolbox.mate(ind1, ind2)
                del ind1.fitness.values
                del ind2.fitness.values
        
        # Note here that we have a different sheme of mutation than in the
        # original algorithm, we use 3 different mutations subsequently.
        for ind in offspring:
            if random.random() < MUTPB:
                toolbox.mutate(ind)
                del ind.fitness.values
            if random.random() < ADDPB:
                toolbox.addwire(ind)
                del ind.fitness.values
            if random.random() < DELPB:
                toolbox.delwire(ind)
                del ind.fitness.values
                
        # Evaluate the individuals with an invalid fitness
        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        evalBegin = time.time()
        fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit
        evaluationTime += (time.time() - evalBegin)
        
        population = toolbox.select(population+offspring, len(offspring))
        hof.update(population)
        stats.update(population)
        logger.logGeneration(gen=g, evals=len(invalid_ind), stats=stats, time=evaluationTime)

    best_network = sn.SortingNetwork(INPUTS, hof[0])
    print(best_network)
    print(best_network.draw())
    print("%i errors, length %i, depth %i" % hof[0].fitness.values)
    totalTime = time.time() - beginTime
    
    print("Total time: {0}\nEvaluation time: {1}".format(totalTime, evaluationTime))
    if args.filename:
        f = open(args.filename, "a")
        f.write("{0};{1};{2};{3}\n".format(args.cores, INPUTS, totalTime, evaluationTime))
        f.close()
        
    return population, stats, hof

if __name__ == "__main__":
    main()



#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
"""
Example of shared constants use.
This is a synthetic partition and evaluation example that should only be
analysed for its shared module API.
"""
from itertools import product
import string
from scoop import futures, shared

# Create the hash to brute force
HASH_TO_FIND = hash("SCO")


def generateHashes(iterator):
    """Compute hashes of given iterator elements"""
    for combination in iterator:
        # Stop as soon as a worker finds the solution
        if shared.getConst('Done', timeout=0):
            return False

        # Compute the current combination hash
        currentString = "".join(combination).strip()
        if hash(currentString) == HASH_TO_FIND:
            # Share to every other worker that the solution has been found
            shared.setConst(Done=True)
            return currentString
    
    # Report that computing has not ended
    return False


if __name__ == "__main__":
    # Generate possible characters
    possibleCharacters = []
    possibleCharacters.extend(list(string.ascii_uppercase))
    possibleCharacters.extend(' ')

    # Generate the solution space.
    stringIterator = product(possibleCharacters, repeat=3)

    # Partition the solution space into iterators
    # Keep in mind that it evaluates the whole solution space
    # making it pretty memory inefficient.
    SplittedIterator = [stringIterator for _ in range(1000)]

    # Parallelize the solution space evaluation
    results = futures.map(generateHashes, SplittedIterator)

    # Loop until a solution is found
    for result in results:
        if result:
            break

    print(result)



#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
"""
Shows the usage of the map_as_completed() function
"""
from scoop import futures


def hello(input_):
    return input_


if __name__ == "__main__":
    print("Execution of map():")
    # Example of how to use a normal map function
    for out in futures.map(hello, range(10)):
        print("Hello from #{}!".format(out))

    print("Execution of map_as_completed():")
    # Example of map_as_completed usage. Note that the results won't necessarily be ordered
    # like the previous 
    for out in futures.map_as_completed(hello, range(10)):
        print("Hello from #{}!".format(out))



#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
"""
A simple example to show some case of exception handling using Scoop. The exception
handling are done in the same way as they are in python. However, exception dealt
by Scoop using python3 are a lot more clear than if python2 was used.
"""
from __future__ import print_function
from scoop import futures

def func0(n):
    # Task submission is asynchronous; It will return immediately.
    task = futures.submit(func1, n)
    # The call blocks here until it gets the result
    result = task.result()
    return result

def func1(n):
    try:
        # The map alone doesn't throw the exception. The exception is raised
        # in the sum which calls the map generator.
        result = sum(futures.map(func2, [i+1 for i in range(n)]))
    except Exception as err:
        # We could do some stuff here
        raise Exception("This exception is normal")
    return result

def func2(n):
    if n > 10:
        # This exception is treated in func1
        raise Exception(10)
    launches = []
    for i in range(n):
        launches.append(futures.submit(func3, i + 1))
    result = futures.as_completed(launches)
    return sum(res.result() for res in result)

def func3(n):
    result = []
    try:
        result = list(futures.map(func4, [i+1 for i in range(n)]))
    except Exception as e:
        # We return what we can
        return e.args[0] + sum(result)
    # No exception was generated
    return sum(result)

def func4(n):
    result = n*n
    if result > 20:
        # This exception is treated in func3
        raise Exception(result)
    return result

def main():
    task = futures.submit(func0, 20)
    # You can wait for a result before continuing computing
    futures.wait([task], return_when=futures.ALL_COMPLETED)
    result = task.result()
    print(result)
    return result

if __name__ == "__main__":
    main()



#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
"""
A very simple example of recursive nested tasks.
Each task maps 2 others tasks, each of these 2 tasks maps 2 others, etc.,
up to RECURSIVITY_DEPTH.
"""

from scoop import futures

RECURSIVITY_DEPTH = 12

def recursiveFunc(level):
    if level == 0:
        return 1
    else:
        args = [level-1] * 2
        s = sum(futures.map(recursiveFunc, args))
        return s

if __name__ == "__main__":
    result = recursiveFunc(RECURSIVITY_DEPTH)
    print("2^{RECURSIVITY_DEPTH} = {result}".format(**locals()))



#!/usr/bin/env python
#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
import sys
import os
import functools
import argparse
import logging

try:
    import psutil
except ImportError:
    psutil = None


if sys.version_info < (3, 3):
    from imp import load_source as importFunction
    FileNotFoundError = IOError
else:
    import importlib.machinery
    importFunction = lambda name, path: importlib.machinery.SourceFileLoader(name, path).load_module()


import scoop
from ..broker.structs import BrokerInfo
from .. import discovery, utils
if sys.version_info < (2, 7):
    import scoop.backports.runpy as runpy
else:
    import runpy


class Bootstrap(object):
    """Set up SCOOP communication links and launches the client module"""
    def __init__(self):
        self.parser = None
        self.args = None
        self.verbose = 1

    def main(self):
        """Bootstrap an arbitrary script.
        If no agruments were passed, use discovery module to search and connect
        to a broker."""
        if self.args is None:
            self.parse()

        self.log = utils.initLogging(self.verbose)

        # Change to the desired directory
        if self.args.workingDirectory:
            os.chdir(self.args.workingDirectory)

        if not self.args.brokerHostname:
            self.log.info("Discovering SCOOP Brokers on network...")
            pools = discovery.Seek()
            if not pools:
                self.log.error("Could not find a SCOOP Broker broadcast.")
                sys.exit(-1)
            self.log.info("Found a broker named {name} on {host} port "
                          "{ports}".format(
                name=pools[0].name,
                host=pools[0].host,
                ports=pools[0].ports,
            ))
            self.args.brokerHostname = pools[0].host
            self.args.taskPort = pools[0].ports[0]
            self.args.metaPort = pools[0].ports[0]
            self.log.debug("Using following addresses:\n{brokerAddress}\n"
                           "{metaAddress}".format(
                                brokerAddress=self.args.brokerAddress,
                                metaAddress=self.args.metaAddress,
                            ))

            self.args.origin = True

        self.setScoop()

        self.run()

    def makeParser(self):
        """Generate the argparse parser object containing the bootloader
           accepted parameters
        """
        self.parser = argparse.ArgumentParser(description='Starts the executable.',
                                              prog=("{0} -m scoop.bootstrap"
                                                    ).format(sys.executable))

        self.parser.add_argument('--origin',
                                 help="To specify that the worker is the origin",
                                 action='store_true')
        self.parser.add_argument('--brokerHostname',
                                 help="The routable hostname of a broker",
                                 default="")
        self.parser.add_argument('--externalBrokerHostname',
                                 help="Externally routable hostname of local "
                                      "worker",
                                 default="")
        self.parser.add_argument('--taskPort',
                                 help="The port of the broker task socket",
                                 type=int)
        self.parser.add_argument('--metaPort',
                                 help="The port of the broker meta socket",
                                 type=int)
        self.parser.add_argument('--size',
                                 help="The size of the worker pool",
                                 type=int,
                                 default=1)
        self.parser.add_argument('--nice',
                                 help="Adjust the niceness of the process",
                                 type=int,
                                 default=0)
        self.parser.add_argument('--debug',
                                 help="Activate the debug",
                                 action='store_true')
        self.parser.add_argument('--profile',
                                 help="Activate the profiler",
                                 action='store_true')
        self.parser.add_argument('--echoGroup',
                                 help="Echo the process Group ID before launch",
                                 action='store_true')
        self.parser.add_argument('--workingDirectory',
                                 help="Set the working directory for the "
                                      "execution",
                                 default=None)
        self.parser.add_argument('--backend',
                                 help="Choice of communication backend",
                                 choices=['ZMQ', 'TCP'],
                                 default='ZMQ')
        self.parser.add_argument('executable',
                                 nargs='?',
                                 help='The executable to start with scoop')
        self.parser.add_argument('args',
                                 nargs=argparse.REMAINDER,
                                 help='The arguments to pass to the executable',
                                 default=[])
        self.parser.add_argument('--verbose', '-v', action='count',
                                 help=("Verbosity level of this launch script"
                                      "(-vv for "
                                      "more)"), default=1)
        self.parser.add_argument('--quiet', '-q', action='store_true',
                                 help="Suppress the output")

    def parse(self):
        """Generate a argparse parser and parse the command-line arguments"""
        if self.parser is None:
            self.makeParser()
        self.args = self.parser.parse_args()
        self.verbose = self.args.verbose if not self.args.quiet else 0

    def setScoop(self):
        """Setup the SCOOP constants."""
        scoop.IS_RUNNING = True
        scoop.IS_ORIGIN = self.args.origin
        scoop.BROKER = BrokerInfo(
            self.args.brokerHostname,
            self.args.taskPort,
            self.args.metaPort,
            self.args.externalBrokerHostname
                if self.args.externalBrokerHostname
                else self.args.brokerHostname,
        )
        scoop.SIZE = self.args.size
        scoop.DEBUG = self.args.debug
        scoop.MAIN_MODULE = self.args.executable
        scoop.CONFIGURATION = {
          'headless': not bool(self.args.executable),
          'backend': self.args.backend,
        }
        scoop.logger = self.log
        if self.args.nice:
            if not psutil:
                scoop.logger.error("psutil not installed.")
                raise ImportError("psutil is needed for nice functionnality.")
            p = psutil.Process(os.getpid())
            p.set_nice(self.args.nice)

        if scoop.DEBUG or self.args.profile:
            from scoop import _debug

    @staticmethod
    def setupEnvironment(self=None):
        """Set the environment (argv, sys.path and module import) of
        scoop.MAIN_MODULE.
        """
        # get the module path in the Python path
        sys.path.append(os.path.dirname(os.path.abspath(scoop.MAIN_MODULE)))

        # Add the user arguments to argv
        sys.argv = sys.argv[:1]
        if self:
            sys.argv += self.args.args

        try:
            user_module = importFunction(
                "SCOOP_WORKER",
                scoop.MAIN_MODULE,
            )
        except FileNotFoundError as e:
            # Could not find file
            sys.stderr.write('{0}\nFile: {1}\nIn path: {2}\n'.format(
                    str(e),
                    scoop.MAIN_MODULE,
                    sys.path[-1],
                )
            )
            sys.stderr.flush()
            sys.exit(-1)

        globs = {}
        try:
            attrlist = user_module.__all__
        except AttributeError:
            attrlist = dir(user_module)
        for attr in attrlist:
            globs[attr] = getattr(user_module, attr)

        if self:
            return globs
        return user_module

    def run(self, globs=None):
        """Import user module and start __main__
           passing globals() is required when subclassing in another module
        """
        # Without this, the underneath import clashes with the top-level one
        global scoop

        if globs is None:
            globs = globals()

        # Show the current process Group ID if asked
        if self.args.echoGroup:
            sys.stdout.write(str(os.getpgrp()) + "\n")
            sys.stdout.flush()

            scoop.logger.info("Worker(s) launched using {0}".format(
                    os.environ['SHELL'],
                )
            )

        # import the user module
        if scoop.MAIN_MODULE:
            globs.update(self.setupEnvironment(self))

        # Start the user program
        from scoop import futures

        def futures_startup():
            """Execute the user code.
            Wraps futures._startup (SCOOP initialisation) over the user module.
            Needs """
            return futures._startup(
                functools.partial(
                    runpy.run_path,
                    scoop.MAIN_MODULE,
                    init_globals=globs,
                    run_name="__main__"
                )
            )

        if self.args.profile:
            import cProfile
            # runctx instead of run is required for local function
            try:
                os.makedirs("profile")
            except:
                pass
            cProfile.runctx(
                "futures_startup()",
                globs,
                locals(),
                "./profile/{0}.prof".format(os.getpid())
            )
        else:
            try:
                futures_startup()
            finally:
                # Must reimport (potentially not there after bootstrap)
                import scoop

                # Ensure a communication queue exists (may happend when a
                # connection wasn't established such as cloud-mode wait).
                if scoop._control.execQueue:
                    scoop._control.execQueue.shutdown()

if __name__ == "__main__":
    b = Bootstrap()
    b.main()



from __future__ import print_function
import random
import math
import time
import argparse
try:
    import cPickle as pickle
except ImportError:
    import pickle

try:
    import pydot
except ImportError:
    pydot = None

GlobalTree = None
mapfunc = None
maxHeight = 0
maxDepth = 0
if pydot:
    g = pydot.Dot()
    h = 0

class Tree():
    def __init__(self, height, minChildren, maxChildren, a, intMean, floatMean):
        global maxHeight, maxDepth, g, h
        import numpy.random
        self.intRange = int(numpy.random.weibull(a)*intMean)
        self.floatRange = int(numpy.random.weibull(a)*floatMean)

        # Record effective height of tree
        if maxHeight == 0:
            maxHeight = height
            maxDepth = height
        elif maxDepth > height:
            maxDepth = height
        self.children = [] 
        self.nodes = 1
        self.leaves = 0
        
        # Generate a new node in graph
        if pydot:
            self.node = pydot.Node( h )
            g.add_node( self.node )
            h += 1
        
        if height == 0:
            self.leaves = 1
            return

        # Generate children
        n = random.randint(minChildren, maxChildren)
        self.children = []
        for i in range(n):
            self.children.append(Tree(height - 1,
                                      minChildren,
                                      maxChildren,
                                      a,
                                      intMean,
                                      floatMean))

        # Keep statistics
        if len(self.children) == 0:
            self.leaves = 1
        
        for child in self.children:
            self.nodes += child.nodes
            self.leaves += child.leaves
            if pydot:
                g.add_edge( pydot.Edge( self.node, child.node ) )
                del child.node
            
        self.height = maxHeight - maxDepth
        if pydot and height == maxHeight:
            del self.node

    def __str__(self):
        if pydot:
            g.write_png("graph.png")
        return ("height : {0}\n"
                "leaves : {1}\n"
                "nodes  : {2}").format(self.height,
                                       self.leaves,
                                       self.nodes)

    def intCalc(self):
        x = 1
        for i in range(self.intRange):
            x = (x + x * i) % 2**32

    def floatCalc(self):
        x = 1.1
        for i in range(self.floatRange):
            x = (x + x * i) % 2**32

def getTree(addresses):
    t = GlobalTree
    for index in addresses:
        t = t.children[index]
    return t

def executeTree(address=[]):
    """This function executes a tree. To limit the size of the arguments passed
    to the function, the tree must be loaded in memory in every worker. To do
    this, simply call "Tree = importTree(filename)" before using the startup
    method of the parralisation library you are using"""
    global nodeDone
    # Get tree subsection
    localTree = getTree(address)
    # Execute tasks
    localTree.intCalc()
    localTree.floatCalc()
    # Select next nodes to be executed
    nextAddresses = [address + [i] for i in range(len(localTree.children))]
    if len(localTree.children) == 0:
        return 1
    # Execute the children
    res = sum(mapfunc(executeTree, nextAddresses))
    assert res == localTree.leaves, (
        "Test failed: res = {0}, leaves = {1}").format(res, localTree.leaves)
    return res

def exportTree(tree, filename):
    f = open(filename, 'wb')
    pickle.dump(tree, f)
    f.close()

def importTree(filename):
    global GlobalTree
    f = open(filename, 'rb')
    GlobalTree = pickle.load(f)
    f.close()

def registerMap(newMap):
    global mapfunc
    mapfunc = newMap    
    
def calibrate(meanTime):
    x = random.randint(1, 9)
    bt = time.time()
    for i in range(1000000):
        x = (x + x * i) % 2**32
    total = time.time() - bt
    intValue = meanTime * 1000000 / total / 2
    
    x = 1.1 * random.randint(1, 9)
    bt = time.time()
    for i in range(1000000):
        x = (x + x * i) % 2**32
    total = time.time() - bt
    floatValue = meanTime * 1000000 / total / 2
    
    print(intValue, floatValue)
    return intValue, floatValue


#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
import marshal
import tempfile
import types
import os
from inspect import ismodule
from functools import partial
try:
    import cPickle as pickle
except ImportError:
    import pickle

try:
    import copyreg
    from io import BytesIO as FileLikeIO
    from io import BufferedReader as FileType
except ImportError:
    # Support for Python 2.X
    import copy_reg as copyreg
    from StringIO import StringIO as FileLikeIO
    from types import FileType as FileType

import scoop


def functionFactory(in_code, name, defaults, globals_, imports):
    """Creates a function at runtime using binary compiled inCode"""
    def generatedFunction():
        pass
    generatedFunction.__code__ = marshal.loads(in_code)
    generatedFunction.__name__ = name
    generatedFunction.__defaults = defaults
    generatedFunction.__globals__.update(pickle.loads(globals_))
    for key, value in imports.items():
        imported_module = __import__(value)
        scoop.logger.debug("Dynamically loaded module {0}".format(value))
        generatedFunction.__globals__.update({key: imported_module})
    return generatedFunction


class FunctionEncapsulation(object):
    """Encapsulates a function in a serializable way.

    This is used by the sharing module (setConst).
    Used for lambda functions and function defined on-the-fly (interactive
    shell)"""
    def __init__(self, in_func, name):
        """Creates a serializable (picklable) object of a function"""
        self.code = marshal.dumps(in_func.__code__)
        self.name = name
        self.defaults = in_func.__defaults__
        # Pickle references to functions used in the function
        used_globals = {} # name: function
        used_modules = {} # used name: origin module name
        for key, value in in_func.__globals__.items():
            if key in in_func.__code__.co_names:
                if ismodule(value):
                    used_modules[key] = value.__name__
                else:
                    used_globals[key] = value
        self.globals = pickle.dumps(used_globals, pickle.HIGHEST_PROTOCOL)
        self.imports = used_modules

    def __call__(self, *args, **kwargs):
        """Called by local worker (which doesn't _communicate this class)"""
        return self.getFunction()(*args, **kwargs)

    def __name__(self):
        return self.name

    def getFunction(self):
        """Called by remote workers. Useful to populate main module globals()
        for interactive shells. Retrieves the serialized function."""
        return functionFactory(
            self.code,
            self.name,
            self.defaults,
            self.globals,
            self.imports,
        )


class ExternalEncapsulation(object):
    """Encapsulates an arbitrary file in a serializable way"""
    def __init__(self, in_filepath):
        """Creates a serializable (picklable) object of inFilePath"""
        self.filename = os.path.basename(in_filepath)
        with open(in_filepath, "rb") as fhdl:
            self.data = pickle.dumps(fhdl, pickle.HIGHEST_PROTOCOL)

    def writeFile(self, directory=None):
        """Writes back the file to a temporary path (optionaly specified)"""
        if directory:
            # If a directory was specified
            full_path = os.path.join(directory, self.filename)
            with open(full_path, 'wb') as f:
                f.write(pickle.loads(self.data).read())
            return full_path

        # if no directory was specified, create a temporary file
        this_file = tempfile.NamedTemporaryFile(delete=False)
        this_file.write(pickle.loads(self.data).read())
        this_file.close()

        return this_file.name


# The following block handles callables pickling and unpickling

# TODO: Make a factory to generate unpickling functions
def unpickleLambda(pickled_callable):
    # TODO: Set globals to user module
    return types.LambdaType(marshal.loads(pickled_callable), globals())

def unpickleMethodType(pickled_callable):
    # TODO: Set globals to user module
    return types.MethodType(marshal.loads(pickled_callable), globals())


def pickleCallable(callable_, unpickle_func):
    # TODO: Pickle also argdefs and closure
    return unpickle_func, (marshal.dumps(callable_.__code__), )

pickle_lambda = partial(pickleCallable, unpickle_func=unpickleLambda)
pickle_method = partial(pickleCallable, unpickle_func=unpickleMethodType)


def makeLambdaPicklable(lambda_function):
    """Take input lambda function l and makes it picklable."""
    if isinstance(lambda_function,
                  type(lambda: None)) and lambda_function.__name__ == '<lambda>':
        def __reduce_ex__(proto):
            # TODO: argdefs, closure
            return unpickleLambda, (marshal.dumps(lambda_function.__code__), )
        lambda_function.__reduce_ex__ = __reduce_ex__
    return lambda_function


# The following block handles file-like objects pickling and unpickling

def unpickleFileLike(position, data):
    file_ = FileLikeIO(data)
    file_.seek(position)
    return file_

def pickleFileLike(file_):
    position = file_.tell()
    file_.seek(0)
    data = file_.read()
    file_.seek(position)
    return unpickleFileLike, (position, data)

copyreg.pickle(FileType, pickleFileLike, unpickleFileLike)



#!/usr/bin/env python
#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
# Global imports
import argparse
import os
import sys
import socket
import subprocess
import time
import logging
import traceback
import signal
from threading import Thread

# Local imports
from scoop import utils
from scoop.launch import Host
from scoop.launch.brokerLaunch import localBroker, remoteBroker
from .broker.structs import BrokerInfo
import scoop

try:
    signal.signal(signal.SIGQUIT, utils.KeyboardInterruptHandler)
except AttributeError:
    # SIGQUIT doesn't exist on Windows
    signal.signal(signal.SIGTERM, utils.KeyboardInterruptHandler)


class ScoopApp(object):
    """SCOOP application. Coordinates the broker and worker launches."""
    LAUNCH_HOST_CLASS = Host

    def __init__(self, hosts, n, b, verbose, python_executable,
            externalHostname, executable, arguments, tunnel, path, debug,
            nice, env, profile, pythonPath, prolog, backend):
        # Assure setup sanity
        assert type(hosts) == list and hosts, ("You should at least "
                                               "specify one host.")
        self.workersLeft = n
        self.createdSubprocesses = []

        # launch information
        self.python_executable = python_executable[0]
        self.pythonpath = pythonPath
        self.prolog = prolog
        self.n = n
        self.b = b
        self.tunnel = tunnel
        self.executable = executable
        self.args = arguments
        self.verbose = verbose
        self.path = path
        self.debug = debug
        self.nice = nice
        self.profile = profile
        self.backend = backend
        self.errors = None

        # Logging configuration
        if self.verbose > 2:
            self.verbose = 2

        scoop.logger = utils.initLogging(
            verbosity=self.verbose,
            name="launcher",
        )

        # Show runtime information (useful for debugging)
        scoop.logger.info("SCOOP {0} {1} on {2} using Python {3}, API: {4}".format(
                scoop.__version__,
                scoop.__revision__,
                sys.platform,
                sys.version.replace("\n", ""),
                sys.api_version,
            )
        )

        if env in ["SLURM","PBS", "SGE"]:
            scoop.logger.info("Detected {0} environment.".format(env))
        scoop.logger.info("Deploying {0} worker(s) over {1} "
                      "host(s).".format(
                          n,
                          len(hosts)
                      )
        )

        # Handling External Hostname
        self.externalHostname = '127.0.0.1' if self.tunnel else externalHostname
        scoop.logger.debug('Using hostname/ip: "{0}" as external broker '
                      'reference.'.format(self.externalHostname))
        scoop.logger.debug('The python executable to execute the program with is: '
                     '{0}.'.format(self.python_executable))

        # Create launch lists
        self.broker_hosts = self.divideHosts(hosts[:], self.b)
        self.worker_hosts = self.divideHosts(hosts, self.n)

        # Logging of worker distribution warnings
        maximumWorkers = sum(host[1] for host in hosts)
        if self.n > maximumWorkers:
            scoop.logger.debug("The -n flag is set at {0} workers, which is higher "
                           "than the maximum number of workers ({1}) specified "
                           "by the hostfile.\nThis behavior may degrade the "
                           "performances of scoop for cpu-bound operations."
                           "".format(qty, maximumWorkers))
        elif self.n < maximumWorkers:
            scoop.logger.debug("The -n flag is set at {0} workers, which is lower "
                           "than the maximum number of workers ({1}) specified "
                           "by the hostfile."
                           "".format(qty, maximumWorkers))

        # Display
        self.showHostDivision(headless=not executable)

        self.workers = []
        self.brokers = []

    def initLogging(self):
        """Configures the logger."""
        verbose_levels = {
            0: logging.WARNING,
            1: logging.INFO,
            2: logging.DEBUG,
        }
        logging.basicConfig(
            level=verbose_levels[self.verbose],
            format="[%(asctime)-15s] %(module)-9s %(levelname)-7s %(message)s"
        )
        return logging.getLogger(self.__class__.__name__)

    def divideHosts(self, hosts, qty):
        """Divide processes among hosts."""
        maximumWorkers = sum(host[1] for host in hosts)

        # If specified amount of workers is greater than sum of each specified.
        if qty > maximumWorkers:
            index = 0
            while qty > maximumWorkers:
                hosts[index] = (hosts[index][0], hosts[index][1] + 1)
                index = (index + 1) % len(hosts)
                maximumWorkers += 1

        # If specified amount of workers if lower than sum of each specified.
        elif qty < maximumWorkers:
            
            while qty < maximumWorkers:
                maximumWorkers -= hosts[-1][1]
                if qty > maximumWorkers:
                    hosts[-1] = (hosts[-1][0], qty - maximumWorkers)
                    maximumWorkers += hosts[-1][1]
                else:
                    del hosts[-1]


        # Checking if the broker if externally routable
        if self.externalHostname in utils.loopbackReferences and \
                len(hosts) > 1 and \
                not self.tunnel:
            raise Exception("\n"
                            "Could not find route from external worker to the "
                            "broker: Unresolvable hostname or IP address.\n "
                            "Please specify your externally routable hostname "
                            "or IP using the --external-hostname parameter or "
                            " use the --tunnel flag.")

        hosts.reverse()
        return hosts

    def showHostDivision(self, headless):
        """Show the worker distribution over the hosts"""
        scoop.logger.info('Worker distribution: ')
        for worker, number in reversed(self.worker_hosts):
            first_worker = (worker == self.worker_hosts[-1][0])
            scoop.logger.info('   {0}:\t{1} {2}'.format(
                worker,
                number - 1 if first_worker or headless else str(number),
                "+ origin" if first_worker or headless else "",
                )
            )

    def _addWorker_args(self, workerinfo):
        """Create the arguments to pass to the addWorker call.
            The returned args and kwargs must ordered/named according to the namedtuple
            in LAUNCH_HOST_CLASS.LAUNCHING_ARGUMENTS

            both args and kwargs are supported for full flexibilty,
            but usage of kwargs only is strongly advised

            workerinfo is a dict with information that can be used to start the worker
        """
        args = []
        kwargs = {
            'pythonPath': self.pythonpath,
            'prolog': self.prolog,
            'path': self.path,
            'nice': self.nice,
            'pythonExecutable': self.python_executable,
            'size': self.n,
            'origin': self.workersLeft == 1,
            'brokerHostname': self.externalHostname,
            'brokerPorts': (self.brokers[0].brokerPort,
                            self.brokers[0].infoPort),
            'debug': self.debug,
            'profiling': self.profile,
            'executable': self.executable,
            'verbose': self.verbose,
            'backend': self.backend,
            'args': self.args,
        }
        return args, kwargs

    def addWorkerToHost(self, workerinfo):
        """Adds a worker to current host"""
        hostname = workerinfo['hostname']

        scoop.logger.debug('Initialising {0}{1} worker {2} [{3}].'.format(
            "local" if hostname in utils.localHostnames else "remote",
            " origin" if self.workersLeft == 1 else "",
            self.workersLeft,
            hostname,
            )
        )

        add_args, add_kwargs = self._addWorker_args(workerinfo)
        self.workers[-1].addWorker(*add_args, **add_kwargs)

    def run(self):
        """Launch the broker(s) and worker(s) assigned on every hosts."""
        # Launch the brokers
        for hostname, nb_brokers in self.broker_hosts:
            for ind in range(nb_brokers):
                # Launching the broker(s)
                if self.externalHostname in utils.localHostnames:
                    self.brokers.append(localBroker(
                        debug=self.debug,
                        nice=self.nice,
                        backend=self.backend,
                    ))
                else:
                    self.brokers.append(remoteBroker(
                        hostname=hostname,
                        pythonExecutable=self.python_executable,
                        debug=self.debug,
                        nice=self.nice,
                        backend=self.backend,
                    ))

        # Share connection information between brokers
        if self.b > 1:
            for broker in self.brokers:
                # Only send data of other brokers to a given broker
                connect_data = [
                    BrokerInfo(
                        x.getHost(),
                        *x.getPorts(),
                        externalHostname=x.getHost()
                    )
                    for x in self.brokers
                    if x is not broker
                ]
                broker.sendConnect(connect_data)

        # Launch the workers
        for hostname, nb_workers in self.worker_hosts:
            self.workers.append(self.LAUNCH_HOST_CLASS(hostname))
            total_workers_host = min(nb_workers, self.workersLeft)
            for worker_idx_host in range(total_workers_host):
                workerinfo = {
                    'hostname': hostname,
                    'total_workers_host': total_workers_host,
                    'worker_idx_host': worker_idx_host,
                }
                self.addWorkerToHost(workerinfo)
                self.workersLeft -= 1

            # Launch every workers at the same time
            scoop.logger.debug(
                "{0}: Launching '{1}'".format(
                    hostname,
                    self.workers[-1].getCommand(),
                )
            )
            shells = self.workers[-1].launch(
                (self.brokers[0].brokerPort,
                 self.brokers[0].infoPort)
                    if self.tunnel else None,
                stdPipe=not self.workers[-1].isLocal(),
            )
            if self.workersLeft <= 0:
                # We've launched every worker we needed, so let's exit the loop
                rootProcess = shells[-1]
                break

        # Wait for the root program
        if self.workers[-1].isLocal():
            self.errors = self.workers[-1].subprocesses[-1].wait()
        else:
            # Process stdout first, then the whole stderr at the end
            for outStream, inStream in [(sys.stdout, rootProcess.stdout),
                                        (sys.stderr, rootProcess.stderr)]:
                data = inStream.read(1)
                while len(data) > 0:
                    # Should not rely on utf-8 codec
                    outStream.write(data.decode("utf-8"))
                    outStream.flush()
                    data = inStream.read(1)
            self.errors = rootProcess.wait()
        scoop.logger.info('Root process is done.')
        return self.errors

    def close(self):
        """Subprocess cleanup."""
        # Give time to flush data if debug was on
        if self.debug:
            time.sleep(5)

        # Terminate workers
        for host in self.workers:
            host.close()

        # Terminate the brokers
        for broker in self.brokers:
            try:
                broker.close()
            except AttributeError:
                # Broker was not started (probably mislaunched)
                pass

        scoop.logger.info('Finished cleaning spawned subprocesses.')


def makeParser():
    """Create the SCOOP module arguments parser."""
    # TODO: Add environment variable (all + selection)
    parser = argparse.ArgumentParser(
        description="Starts a parallel program using SCOOP.",
        prog="{0} -m scoop".format(sys.executable),
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--hosts', '--host',
                       help="The list of hosts. The first host will execute "
                            "the origin. (default is 127.0.0.1)",
                       metavar="Address",
                       nargs='*')
    group.add_argument('--hostfile',
                       help="The hostfile name",
                       metavar="FileName")
    parser.add_argument('--path', '-p',
                        help="The path to the executable on remote hosts "
                             "(default is local directory)",
                        default=os.getcwd())
    parser.add_argument('--nice',
                        type=int,
                        metavar="NiceLevel",
                        help="*nix niceness level (-20 to 19) to run the "
                             "executable")
    parser.add_argument('--verbose', '-v',
                        action='count',
                        help="Verbosity level of this launch script (-vv for "
                             "more)",
                        default=1)
    parser.add_argument('--quiet', '-q',
                        action='store_true')
    parser.add_argument('-n',
                        help="Total number of workers to launch on the hosts. "
                             "Workers are spawned sequentially over the hosts. "
                             "(ie. -n 3 with 2 hosts will spawn 2 workers on "
                             "the first host and 1 on the second.) (default: "
                             "Number of CPUs on current machine)",
                        type=int,
                        metavar="NumberOfWorkers")
    parser.add_argument('-b',
                        help="Total number of brokers to launch on the hosts. "
                             "Brokers are spawned sequentially over the hosts. "
                             "(ie. -b 3 with 2 hosts will spawn 2 brokers on "
                             "the first host and 1 on the second.) (default: "
                             "1)",
                        type=int,
                        default=1,
                        metavar="NumberOfBrokers")
    parser.add_argument('--tunnel',
                        help="Activate ssh tunnels to route toward the broker "
                             "sockets over remote connections (may eliminate "
                             "routing problems and activate encryption but "
                             "slows down communications)",
                        action='store_true')
    parser.add_argument('--external-hostname',
                        nargs=1,
                        help="The externally routable hostname / ip of this "
                             "machine. (defaults to the local hostname)",
                        metavar="Address")
    parser.add_argument('--python-interpreter',
                        nargs=1,
                        help="The python interpreter executable with which to "
                             "execute the script",
                        default=[sys.executable],
                        metavar="Path")
    parser.add_argument('--pythonpath',
                        nargs=1,
                        help="The PYTHONPATH environment variable (default is "
                             "current PYTHONPATH)",
                        default=[os.environ.get('PYTHONPATH', '')])
    parser.add_argument('--prolog',
                        nargs=1,
                        help="Absolute Path to a shell script or executable "
                             "that will be executed at the launch of every "
                             "worker",
                        default=[None])
    parser.add_argument('--debug',
                        help=argparse.SUPPRESS,
                        action='store_true')
    parser.add_argument('--profile',
                        help=("Turn on the profiling. SCOOP will call "
                        "cProfile.run on the executable for every worker and"
                        " will produce files in directory profile/ named "
                        "workerX where X is the number of the worker."),
                        action='store_true')
    parser.add_argument('--backend',
                        help="Choice of communication backend",
                        choices=['ZMQ', 'TCP'],
                        default='ZMQ')
    parser.add_argument('executable',
                        nargs='?',
                        help='The executable to start with SCOOP')
    parser.add_argument('args',
                        nargs=argparse.REMAINDER,
                        help='The arguments to pass to the executable',
                        default=[],
                        metavar="args")
    return parser


def main():
    """Execution of the SCOOP module. Parses its command-line arguments and
    launch needed resources."""
    # Generate a argparse parser and parse the command-line arguments
    parser = makeParser()
    args = parser.parse_args()

    # Get a list of resources to launch worker(s) on
    hosts = utils.getHosts(args.hostfile, args.hosts)
    if args.n:
        n = args.n
    else:
        n = utils.getWorkerQte(hosts)
    assert n > 0, ("Scoop couldn't determine the number of worker to start.\n"
                   "Use the '-n' flag to set it manually.")
    if not args.external_hostname:
        args.external_hostname = [utils.externalHostname(hosts)]

    # Launch SCOOP
    thisScoopApp = ScoopApp(hosts, n, args.b,
                            args.verbose if not args.quiet else 0,
                            args.python_interpreter,
                            args.external_hostname[0],
                            args.executable, args.args, args.tunnel,
                            args.path, args.debug, args.nice,
                            utils.getEnv(), args.profile, args.pythonpath[0],
                            args.prolog[0], args.backend)

    rootTaskExitCode = False
    interruptPreventer = Thread(target=thisScoopApp.close)
    try:
        rootTaskExitCode = thisScoopApp.run()
    except Exception as e:
        logging.error('Error while launching SCOOP subprocesses:')
        logging.error(traceback.format_exc())
        rootTaskExitCode = -1
    finally:
        # This should not be interrupted (ie. by a KeyboadInterrupt)
        # The only cross-platform way to do it I found was by using a thread.
        interruptPreventer.start()
        interruptPreventer.join()

    # Exit with the proper exit code
    if rootTaskExitCode:
        sys.exit(rootTaskExitCode)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from Tree import *
import sys
import time
from deap import dtm

def main():
    return executeTree()
    
importTree(sys.argv[1] if len(sys.argv) > 1 else "tree.txt")
registerMap(dtm.map)

if __name__=="__main__":
    bt = time.time()
    dtm.start(main)
    totalTime = time.time() - bt
    print("total time : {}".format(totalTime))